from mage_ai.data_cleaner.transformer_actions.udf.base import BaseUDF


class StringReplace(BaseUDF):
    def execute(self):
        pattern = self.options.get('pattern')
        replacement = self.options.get('replacement')
        if not pattern and not replacement:
            raise Exception(f'Require both `pattern` and `replacement` parameters.')
        return self.df[self.arguments[0]].str.replace(pattern, replacement)



import os
import re


def clean_name(name):
    for c in ['\ufeff', '\uFEFF', '"', '$', '\n', '\r', '\t']:
        name = name.replace(c, '')
    name = re.sub('\W', '_', name)

    if name and re.match('\d', name[0]):
        name = f'letter_{name}'
    return name.lower()


def files_in_path(path, verbose=0):
    files = []
    # r=root, d=directories, f = files
    for r, d, f in os.walk(path):
        for file in f:
            files.append(os.path.join(r, file))

    if verbose >= 1:
        for f in files:
            print(f)

    return files


def files_in_single_path(path):
    f = []
    for (dirpath, dirnames, filenames) in os.walk(path):
        f.extend([os.path.join(dirpath, file) for file in filenames])
        break
    return f



from mage_ai.data_cleaner.pipelines.base import BasePipeline
from mage_ai.server.client.mage import Mage
from mage_ai.server.data.base import Model
from mage_ai.shared.constants import SAMPLE_SIZE
from mage_ai.shared.hash import merge_dict
import logging
import os
import os.path

logger = logging.getLogger(__name__)

# right now, we are writing the models to local files to reduce dependencies
class FeatureSet(Model):
    def __init__(self, id=None, df=None, name=None, api_key=None):
        super().__init__(id)

        # Update metadata
        try:
            metadata = self.metadata
        except Exception:
            self.metadata = {}
            metadata = self.metadata

        metadata_updated = False
        if name is not None:
            metadata['name'] = name
            metadata_updated = True

        if self.pipeline is None:
            pipeline = Pipeline(feature_set_id=self.id, api_key=api_key)
            metadata['pipeline_id'] = pipeline.id
            metadata_updated = True

        if metadata_updated:
            self.metadata = metadata

        if df is None:
            self._data = None
            self._data_orig = None
        else:
            self.data = df
            self.data_orig = df

    def __repr__(self):
        formatted_suggestions = []
        for idx, s in enumerate(self.suggestions):
            formatted_suggestions.append(
                f'{idx + 1}. {s["title"]}({s["action_payload"]["action_arguments"]}):'
                f' {s["message"]}',
            )
        suggestion_str = '\n'.join(formatted_suggestions)
        return f'<FeatureSet {self.id}> Cleaning suggestions:\n{suggestion_str}'

    @property
    def data(self):
        if self._data is None:
            self._data = self.read_parquet_file('data.parquet')
        return self._data

    @property
    def cleaning_rule_configs(self):
        return self.read_json_file('cleaning_rule_configs.json', {})

    @cleaning_rule_configs.setter
    def cleaning_rule_configs(self, configs):
        return self.write_json_file('cleaning_rule_configs.json', configs)

    @data.setter
    def data(self, df):
        self.write_parquet_file(os.path.join(self.dir, 'data.parquet'), df)
        self._data = df

    @property
    def data_orig(self):
        if self._data_orig is None:
            self._data_orig = self.read_parquet_file('data_orig.parquet')
        return self._data_orig

    @data_orig.setter
    def data_orig(self, df):
        self.write_parquet_file(os.path.join(self.dir, 'data_orig.parquet'), df)
        self._data_orig = df

    @property
    def metadata(self):
        return self.read_json_file('metadata.json', {})

    @metadata.setter
    def metadata(self, metadata):
        return self.write_json_file('metadata.json', metadata)

    @property
    def statistics(self):
        return self.read_json_file('statistics.json', {})

    @statistics.setter
    def statistics(self, metadata):
        return self.write_json_file('statistics.json', metadata)

    @property
    def insights(self):
        return self.read_json_file('insights.json', [])

    @insights.setter
    def insights(self, metadata):
        return self.write_json_file('insights.json', metadata)

    @property
    def suggestions(self):
        return self.read_json_file('suggestions.json', [])

    @suggestions.setter
    def suggestions(self, metadata):
        return self.write_json_file('suggestions.json', metadata)

    @property
    def pipeline(self):
        pipeline_id = self.metadata.get('pipeline_id')
        if pipeline_id is None:
            return None
        return Pipeline(id=pipeline_id)

    @property
    def sample_data(self):
        sample_size = SAMPLE_SIZE
        if self.data.size < sample_size:
            sample_size = len(self.data)
        return self.data.head(sample_size)

    def version_snapshot(self, version):
        return self.read_json_file(f'versions/{version}.json', {})

    def write_files(self, obj, write_orig_data=False, prev_version=None):
        if prev_version is not None:
            # Save the old version as a snapshot
            self.write_version_snapshot(prev_version)
        if 'cleaning_rule_configs' in obj:
            self.cleaning_rule_configs = obj['cleaning_rule_configs']
        if 'df' in obj:
            self.data = obj['df']
        if 'metadata' in obj:
            self.metadata = obj['metadata']
        if 'suggestions' in obj:
            self.suggestions = obj['suggestions']
        if 'statistics' in obj:
            self.statistics = obj['statistics']
        if 'insights' in obj:
            self.insights = obj['insights']
        if write_orig_data and 'df' in obj:
            self.data_orig = obj['df']
        # Update metadata
        metadata = self.metadata
        if 'column_types' in obj:
            metadata['column_types'] = obj['column_types']
        if 'statistics' in obj:
            metadata['statistics'] = dict(
                count=obj['statistics']['count'],
                quality='Good' if obj['statistics']['validity'] >= 0.8 else 'Bad',
            )
        self.metadata = metadata

    def write_version_snapshot(self, version):
        version_snapshot = self.to_dict()
        try:
            self.write_json_file(f'{version}.json', version_snapshot, subdir='versions')
        except Exception:
            logger.exception(f'Failed to write snapshot v{version} for feature set {self.id}')
        return version_snapshot

    def to_dict(self, column=None, detailed=True, version=None):
        if version is not None:
            return self.version_snapshot(version)
        obj = dict(
            id=self.id,
            metadata=self.metadata,
        )
        if detailed:
            sample_data = self.sample_data
            datetime_cols = [
                col for col in sample_data.columns if 'datetime64' in str(sample_data[col])
            ]
            sample_data[datetime_cols] = sample_data[datetime_cols].astype(str)
            # Filter sample data
            if column is not None:
                sample_data_dict = dict(
                    columns=[column],
                    rows=[[v] for v in sample_data[column].to_numpy().tolist()],
                )
            else:
                sample_data_dict = dict(
                    columns=sample_data.columns.tolist(),
                    rows=sample_data.to_numpy().tolist(),
                )
            # Filter suggestions
            suggestions = self.suggestions
            if column is not None:
                suggestions = [
                    s for s in suggestions if column in s['action_payload']['action_arguments']
                ]
                for s in suggestions:
                    s['action_payload']['action_arguments'] = [column]

            # Deduplicate outlier removal suggestions
            pipeline_dict = self.pipeline.to_dict()
            statistics = self.statistics
            suggestions_filtered, statistics_updated = BasePipeline.deduplicate_suggestions(
                pipeline_dict['actions'],
                suggestions,
                statistics,
            )
            obj = merge_dict(
                obj,
                dict(
                    pipeline=pipeline_dict,
                    sample_data=sample_data_dict,
                    statistics=statistics_updated,
                    insights=self.insights,
                    cleaning_rule_configs=self.cleaning_rule_configs,
                    suggestions=suggestions_filtered,
                ),
            )
        return obj


class Pipeline(Model):
    def __init__(self, id=None, path=None, feature_set_id=None, pipeline=None, api_key=None):
        super().__init__(id=id, path=path)

        # Update metadata
        try:
            metadata = self.metadata
        except Exception:
            self.metadata = {}
            metadata = self.metadata

        if feature_set_id is not None:
            metadata['feature_set_id'] = int(feature_set_id)
            self.metadata = metadata

        if pipeline is not None:
            self.pipeline = pipeline
            if api_key is not None:
                self.sync_pipeline(api_key)

    def sync_pipeline(self, api_key):
        Mage().sync_pipeline(self, api_key)

    @property
    def metadata(self):
        return self.read_json_file('metadata.json', {})

    @metadata.setter
    def metadata(self, metadata):
        return self.write_json_file('metadata.json', metadata)

    @property
    def pipeline(self):
        fpath = 'pipeline.json'
        if self.path is not None and os.path.isfile(self.path):
            fpath = self.path
        actions = self.read_json_file(fpath, [])
        return BasePipeline(actions=actions)

    @pipeline.setter
    def pipeline(self, pipeline):
        return self.write_json_file('pipeline.json', pipeline.actions)

    def get_feature_set(self):
        feature_set_id = self.metadata.get('feature_set_id')
        if feature_set_id is not None:
            return FeatureSet(id=feature_set_id)

    def to_dict(self, detailed=True):
        return dict(
            id=self.id,
            metadata=self.metadata,
            actions=self.pipeline.actions,
        )



from mage_ai.data_preparation.repo_manager import get_repo_path
from typing import Dict
import os

BLACKLISTED_DIRS = frozenset(['venv', 'env', '.git', '.DS_Store', '__pycache__'])
INACCESSIBLE_DIRS = frozenset(['__pycache__'])
MAX_DEPTH = 30


class File:
    def __init__(
        self,
        filename: str,
        dir_path: str,
        repo_path: str,
    ) -> None:
        self.filename = filename
        self.dir_path = dir_path
        self.repo_path = repo_path

    @property
    def file_path(self) -> str:
        return os.path.join(self.repo_path, self.dir_path, self.filename)

    @classmethod
    def create(self, filename, dir_path, repo_path=None):
        repo_path = repo_path or get_repo_path()
        file = File(filename, dir_path, repo_path)
        with open(file.file_path, 'w'):
            pass
        return file

    @classmethod
    def from_path(self, file_path, repo_path=None):
        repo_path = repo_path or get_repo_path()
        return File(os.path.basename(file_path), os.path.dirname(file_path), repo_path)

    @classmethod
    def get_all_files(self, repo_path):
        return traverse(os.path.basename(repo_path), True, repo_path)

    def content(self):
        with open(self.file_path) as fp:
            file_content = fp.read()
        return file_content

    def update_content(self, content):
        with open(self.file_path, 'w') as fp:
            fp.write(content)

    def delete(self):
        os.remove(self.file_path)

    def to_dict(self, include_content=False):
        data = dict(name=self.filename, path=os.path.join(self.dir_path, self.filename))
        if include_content:
            data['content'] = self.content()
        return data


def traverse(name: str, is_dir: str, path: str, disabled=False, depth=1) -> Dict:
    tree_entry = dict(name=name)
    if not is_dir:
        tree_entry['disabled'] = disabled
        return tree_entry
    if depth >= MAX_DEPTH:
        return tree_entry
    can_access_children = name[0] == '.' or name in INACCESSIBLE_DIRS
    tree_entry['children'] = list(
        traverse(
            entry.name,
            entry.is_dir(follow_symlinks=False),
            entry.path,
            can_access_children,
            depth + 1,
        )
        for entry in sorted(
            filter(lambda entry: entry.name not in BLACKLISTED_DIRS, os.scandir(path)),
            key=lambda entry: entry.name,
        )
    )
    return tree_entry



from botocore.config import Config
from botocore.exceptions import ClientError
from datetime import datetime
from mage_ai.services.emr import emr_basics
from mage_ai.services.emr.config import EmrConfig
import boto3
import logging
import random
import time
import os
import sys


MAX_STEPS_IN_CLUSTER = 255 - 55
MAX_RUNNING_OR_PENDING_STEPS = 15

logger = logging.getLogger(__name__)


def get_running_cluster_count(emr_client):
    clusters = emr_client.list_clusters(
        ClusterStates=[
            'BOOTSTRAPPING',
            'RUNNING',
            'STARTING',
            'WAITING',
            'TERMINATING',
        ]
    )
    return len(clusters['Clusters'])


def create_a_new_cluster(
    cluster_name,
    steps,
    emr_config,
    bootstrap_script_path=None,
    idle_timeout=0,
    log_uri=None,
):
    region_name = os.getenv('AWS_REGION_NAME', 'us-west-2')
    config = Config(region_name=region_name)
    emr_client = boto3.client('emr', config=config)

    print('Creating cluster...')

    applications = ['Hadoop', 'Hive', 'Spark']
    response = emr_client.run_job_flow(
        Name=f'{datetime.utcnow().isoformat()}-{cluster_name}',
        LogUri=log_uri,
        ReleaseLabel='emr-6.5.0',
        Instances=emr_config.get_instances_config(
            get_running_cluster_count(emr_client),
            idle_timeout,
        ),
        Steps=__build_steps_config(steps),
        StepConcurrencyLevel=256,
        Applications=[{
            'Name': app
        } for app in applications],
        BootstrapActions=[
            dict(
                Name='Install Python packages using pip.',
                ScriptBootstrapAction=dict(
                    Path=bootstrap_script_path,
                ),
            ),
        ],
        JobFlowRole='EMR_EC2_DefaultRole',
        ServiceRole='EMR_DefaultRole',
        EbsRootVolumeSize=10,
        VisibleToAllUsers=True,
    )
    print(response)
    print('\n')
    cluster_id = response['JobFlowId']
    print(f'Cluster ID: {cluster_id}')

    # There are two levels of auto-termination:
    #     * Instance level
    #     * Cluster level
    # For instance level settings, we already set it to be auto-terminated, i.e.,
    #
    #     'KeepJobFlowAliveWhenNoSteps': False,
    #
    # which over-writes cluster-level settings at present. For instance level settings,
    # there were hanging instances if we don’t set it to be auto-terminate somehow. Thus
    # the cluster-level AutoTerminationPolicy didn’t take effects so far, and would be
    # commented out for now
    #
    # if idle_timeout > 0:
    #    emr_client.put_auto_termination_policy(
    #        ClusterId=cluster_id,
    #        AutoTerminationPolicy={
    #            'IdleTimeout': idle_timeout
    #        }
    #    )

    __status_poller(
        'Waiting for cluster, this typically takes several minutes...',
        'RUNNING',
        lambda: emr_basics.describe_cluster(cluster_id, emr_client)['Status']['State'],
    )

    return cluster_id


def submit_spark_job(
    cluster_name,
    steps,
    bootstrap_script_path=None,
    emr_config=dict(),
    idle_timeout=0,
    log_uri=None,
):
    region_name = os.getenv('AWS_REGION_NAME', 'us-west-2')
    config = Config(region_name=region_name)
    emr_client = boto3.client('emr', config=config)

    clusters = emr_client.list_clusters(
        ClusterStates=[
            'RUNNING',
            'WAITING',
        ],
    )

    # terminate_cluster_ids = []
    valid_cluster_ids = []
    # clusters_with_steps = {}

    clusters = sorted(clusters['Clusters'], key=lambda x: 0 if x['Status'] == 'WAITING' else 1)

    print('Clusters:')
    for cluster in clusters:
        cluster_id = cluster['Id']
        print(f'\t{cluster_id}')
        current_steps = emr_basics.list_steps(cluster_id, emr_client)
        total_steps = len(current_steps)
        running_steps = len(
            list(
                filter(lambda x: x.get('Status', {}).get('State', {}) == 'RUNNING', current_steps),
            ),
        )
        pending_steps = len(
            list(
                filter(lambda x: x.get('Status', {}).get('State', {}) == 'PENDING', current_steps),
            ),
        )
        active_steps = running_steps + pending_steps

        # clusters_with_steps[cluster_id] = {
        #     'cluster': cluster,
        #     'running_steps': running_steps,
        #     'steps': current_steps,
        #     'total_steps': total_steps,
        # }
        # if total_steps >= MAX_STEPS_IN_CLUSTER and running_steps == 0:
        #     terminate_cluster_ids.append(cluster_id)

        if active_steps < MAX_RUNNING_OR_PENDING_STEPS and total_steps < MAX_STEPS_IN_CLUSTER:
            valid_cluster_ids.append((total_steps, active_steps, cluster_id))

    # if len(terminate_cluster_ids) >= 1:
    #     print('Terminating clusters: {}'.format(terminate_cluster_ids))
    #     response = emr_client.terminate_job_flows(
    #         JobFlowIds=terminate_cluster_ids,
    #     )
    #     print(response)
    #     print('\n')

    if len(valid_cluster_ids) == 0:
        cluster_id = create_a_new_cluster(
            cluster_name,
            steps,
            EmrConfig(config=emr_config),
            bootstrap_script_path=bootstrap_script_path,
            idle_timeout=idle_timeout,
            log_uri=log_uri,
        )

        # for step in steps:
        #     def _get_status():
        #         filtered_steps = list(filter(
        #             lambda x: x['Name'] == step['name'],
        #             emr_basics.list_steps(cluster_id, emr_client),
        #         ))
        #         return filtered_steps[0]['Status']['State']

        #     step_name = step['name']
        #     status = __status_poller(
        #         f'Waiting for step {step_name} to complete...',
        #         'COMPLETED',
        #         _get_status,
        #     )

        #     if status != 'COMPLETED':
        #         raise Exception(f'Cluster ID {cluster_id} did not complete step, status: {status}')
    else:
        total_steps, active_steps, cluster_id = sorted(valid_cluster_ids, key=lambda t: t[0])[0]
        print(f'Number of steps in cluster ID {cluster_id}: {total_steps}')
        print(f'\tActive steps: {active_steps}')

        response, step_id, status = __add_step(
            cluster_id=cluster_id,
            emr_client=emr_client,
            steps=steps,
        )

        if status != 'COMPLETED':
            raise Exception(f'Step ID {step_id} did not complete, status: {status}')


def __add_step(emr_client, cluster_id, steps):
    response = emr_client.add_job_flow_steps(
        JobFlowId=cluster_id,
        Steps=__build_steps_config(steps),
    )

    print(response)
    print('Step IDs:')
    for step_id in response['StepIds']:
        print(f'\t{step_id}')
    print('\n')
    step_id = response['StepIds'][0]

    def _get_status():
        return list(
            filter(
                lambda x: x['Id'] == step_id,
                emr_basics.list_steps(cluster_id, emr_client),
            ),
        )[0]['Status']['State']

    __status_poller(
        f'Waiting for step {step_id} to complete...',
        'COMPLETED',
        _get_status,
    )

    return response, step_id, _get_status()


def __build_steps_config(steps):
    return [{
        'Name': step['name'],
        'ActionOnFailure': 'CONTINUE',
        'HadoopJarStep': {
            'Jar': 'command-runner.jar',
            'Args': [
                'spark-submit',
                '--deploy-mode',
                'cluster',
                step['script_uri'],
                *step['script_args'],
            ]
        }
    } for step in steps]


def __status_poller(intro, done_status, func):
    emr_basics.logger.setLevel(logging.WARNING)
    status = None
    print(intro)
    print("Current status: ", end='')
    statuses = [
        'CANCELLED',
        'FAILED',
        'INTERRUPTED',
        done_status,
    ]

    base_sleep_time = 60
    tries = 0

    random.seed()
    time.sleep(random.randrange(base_sleep_time))
    while status not in statuses:
        tries += 1
        prev_status = status
        try:
            status = func()
            sleep_time = base_sleep_time
        except ClientError as err:
            if 'ListSteps' in str(err) and 'Rate exceeded' in str(err):
                status = prev_status
                sleep_time = base_sleep_time + (30 * tries)
            else:
                raise err

        if prev_status == status:
            print('.', end='')
        else:
            print(status, end='')
        sys.stdout.flush()
        time.sleep(sleep_time)
    print()
    emr_basics.logger.setLevel(logging.INFO)
    return status



from mage_ai.data_cleaner.cleaning_rules.base import BaseRule
from mage_ai.data_cleaner.column_types.column_type_detector import REGEX_NUMBER
from mage_ai.data_cleaner.transformer_actions.constants import ActionType, NameConventionPatterns
from keyword import iskeyword


class CleanColumnNames(BaseRule):
    def is_dirty(self, name):
        if NameConventionPatterns.NON_ALNUM.search(name):
            return True
        if REGEX_NUMBER.match(name) != None:
            return True

        if iskeyword(name):
            return True
        if NameConventionPatterns.SNAKE.match(name):
            return False
        if NameConventionPatterns.LOWERCASE.match(name):
            return False
        return True

    def evaluate(self):
        """
        Rule:
        1. If column name contains an nonalphanumeric character (except _),
           suggest cleaning (remove all characters)
        2. If the column name is a number, suggest cleaning (prefix with 'number_')
        3. If the column name is a reseved keyword, suggest cleaning (postfix with '_')
        4. If the column name is not snake_case, suggest cleaning
           (convert to snakecase from pascal case, camel case, uppercase)
        """
        matches = list(filter(self.is_dirty, self.df_columns))
        suggestions = []
        if len(matches) != 0:
            suggestions.append(
                self._build_transformer_action_suggestion(
                    'Clean dirty column names',
                    'Format these column names as lowercase and alphanumeric to improve '
                    'accessibility and security.',
                    action_type=ActionType.CLEAN_COLUMN_NAME,
                    action_arguments=matches,
                    axis='column',
                )
            )

        return suggestions



from mage_ai.io.base import BaseFile, FileFormat
from pandas import DataFrame
from typing import Union


class FileIO(BaseFile):
    """
    Handles data transfer between the filesystem and the Mage app.
    """

    def load(
        self,
        filepath: str,
        format: FileFormat = None,
        **kwargs,
    ) -> DataFrame:
        """
        Loads the data frame from the filepath specified.

        Args:
            filepath (os.PathLike): Filepath to load data frame from.
            format (Union[FileFormat, str], Optional): Format of the file to load data frame from.
            Defaults to None, in which case the format is inferred.

        Returns:
            DataFrame: Data frame object loaded from the specified data frame.
        """
        if format is None:
            format = self._get_file_format(filepath)
        with self.printer.print_msg(f'Loading data frame from \'{filepath}\''):
            return self._read(filepath, format, **kwargs)

    def export(
        self, df: DataFrame, filepath: str, format: Union[FileFormat, str] = None, **kwargs
    ) -> None:
        """
        Exports the input dataframe to the file specified.

        Args:
            df (DataFrame): Data frame to export.
            filepath (os.PathLike): Filepath to export data frame to.
            format (Union[FileFormat, str], Optional): Format of the file to export data frame to.
            Defaults to None, in which case the format is inferred.
        """
        if format is None:
            format = self._get_file_format(filepath)
        with self.printer.print_msg(f'Exporting data frame to \'{filepath}\''):
            self._write(df, format, filepath, **kwargs)



from contextlib import contextmanager
from typing import Callable

import logging
import time

logger = logging.getLogger(__name__)


class timer(object):
    """
    with timer('metric.metric', tags={ 'key': 'value' }):
        function()
    """

    def __init__(self, metric, tags={}, verbose=True):
        self.metric = metric
        self.start = None
        self.tags = tags
        self.verbose = verbose

    def __enter__(self):
        self.start = time.time()

    def __exit__(self, type, value, traceback):
        # Must convert to milliseconds, see details in
        # https://statsd.readthedocs.io/en/v3.1/timing.html
        dt = int((time.time() - self.start) * 1000)
        if self.verbose:
            logger.debug(f'[time] metric: {self.metric}, value: {dt}ms, tags: {self.tags}')


class VerboseFunctionExec:
    def __init__(
        self,
        message: str,
        log_func: Callable[[str], None] = None,
        prefix: str = '',
        verbose: bool = True,
    ):
        self.message = message
        self.verbose = verbose
        self.prefix = prefix
        self.log_func = log_func

    def __enter__(self):
        if self.verbose:
            enter_message = f'{self.prefix} {self.message}...'
            if self.log_func is None:
                print_kwargs=dict()
                if self.prefix is None or len(self.prefix) == 0:
                    print_kwargs['end'] = ''
                print(enter_message, **print_kwargs)
            else:
                self.log_func(enter_message)

    def __exit__(self, exc_type, exc_value, exc_tb):
        if self.verbose:
            log_func = self.log_func or print
            if exc_type is None:
                log_func(f'{self.prefix} DONE')
            else:
                log_func(f'{self.prefix} FAILED')


class VerbosePrintHandler:
    def __init__(self, start_msg, verbose=False):
        self.verbose = verbose
        if verbose:
            print(start_msg)
        self.exists_previous_message = False

    @contextmanager
    def print_msg(self, msg):
        if self.verbose:
            if self.exists_previous_message:
                print(f'\r├─ ')
            print(f'└─ {msg}...', end='')
        yield msg
        if self.verbose:
            print('DONE', end='')
            self.exists_previous_message = True



from mage_ai.data_preparation.models.constants import (
    BlockType,
    CUSTOM_EXECUTION_BLOCK_TYPES,
)
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.repo_manager import get_repo_config, get_repo_path
from mage_ai.server.active_kernel import (
    get_active_kernel_client,
    get_active_kernel_name,
    switch_active_kernel,
)
from mage_ai.server.kernel_output_parser import DataType
from mage_ai.server.kernels import KernelName
from mage_ai.server.utils.output_display import (
    add_internal_output_info,
    add_execution_code,
    get_block_output_process_code,
    get_pipeline_execution_code,
)
from mage_ai.shared.hash import merge_dict
from jupyter_client import KernelClient
import asyncio
import json
import threading
import tornado.websocket
import traceback
import uuid


class WebSocketServer(tornado.websocket.WebSocketHandler):
    """Simple WebSocket handler to serve clients."""

    # Note that `clients` is a class variable and `send_message` is a
    # classmethod.
    clients = set()
    running_executions_mapping = dict()

    def open(self):
        WebSocketServer.clients.add(self)

    def on_close(self):
        WebSocketServer.clients.remove(self)

    def check_origin(self, origin):
        return True

    def init_kernel_client(self, kernel_name) -> KernelClient:
        if kernel_name != get_active_kernel_name():
            switch_active_kernel(kernel_name)

        return get_active_kernel_client()

    def on_message(self, raw_message):
        message = json.loads(raw_message)
        custom_code = message.get('code')
        output = message.get('output')
        if output:
            self.send_message(output)
            return
        global_vars = message.get('global_vars')
        execute_pipeline = message.get('execute_pipeline')
        kernel_name = message.get('kernel_name', get_active_kernel_name())

        run_downstream = message.get('run_downstream')
        run_upstream = message.get('run_upstream')
        run_tests = message.get('run_tests')
        block_type = message.get('type')
        block_uuid = message.get('uuid')
        pipeline_uuid = message.get('pipeline_uuid')
        pipeline = Pipeline(pipeline_uuid, get_repo_path())


        value = dict(
            block_uuid=block_uuid,
            pipeline_uuid=pipeline_uuid,
        )

        def publish_message(
            message: str,
            execution_state: str = 'busy',
            msg_type: str = 'stream_pipeline',
        ) -> None:
            msg_id = str(uuid.uuid4())
            WebSocketServer.running_executions_mapping[msg_id] = value
            self.send_message(
                dict(
                    data=message,
                    execution_state=execution_state,
                    msg_id=msg_id,
                    msg_type=msg_type,
                    type=DataType.TEXT_PLAIN,
                )
            )

        if execute_pipeline:
            if kernel_name == KernelName.PYSPARK:
                code = get_pipeline_execution_code(
                    pipeline_uuid,
                    global_vars=global_vars,
                    kernel_name=kernel_name,
                    pipeline_config=pipeline.to_dict(include_content=True),
                    repo_config=get_repo_config().to_dict(remote=True),
                    update_status=False if kernel_name == KernelName.PYSPARK else True,
                )
                client = self.init_kernel_client(kernel_name)
                msg_id = client.execute(code)

                WebSocketServer.running_executions_mapping[msg_id] = value
            else:
                def run_pipeline() -> None:
                    try:
                        asyncio.run(pipeline.execute(log_func=publish_message, redirect_outputs=True))
                        publish_message(f'Pipeline {pipeline.uuid} execution complete.', 'idle')
                    except Exception:
                        trace = traceback.format_exc().splitlines()
                        publish_message(f'Pipeline {pipeline.uuid} execution failed with error:')
                        publish_message(trace, 'idle')

                threading.Thread(target=run_pipeline).start()
        else:
            widget = BlockType.CHART == block_type

            block = pipeline.get_block(block_uuid, widget=widget)
            code = custom_code

            client = self.init_kernel_client(kernel_name)

            if not custom_code and BlockType.SCRATCHPAD == block_type:
                msg_id = client.execute('')

                value = dict(
                    block_uuid=block_uuid,
                    pipeline_uuid=pipeline_uuid,
                )
                WebSocketServer.running_executions_mapping[msg_id] = value

                self.send_message(
                    dict(
                        data='',
                        execution_state='idle',
                        msg_id=msg_id,
                        type=DataType.TEXT_PLAIN,
                    ),
                )
            else:
                if block is not None and block.type in CUSTOM_EXECUTION_BLOCK_TYPES:
                    if kernel_name == KernelName.PYSPARK and not widget:
                        remote_execution = True
                    else:
                        remote_execution = False
                    code = add_execution_code(
                        pipeline_uuid,
                        block_uuid,
                        custom_code,
                        global_vars,
                        analyze_outputs=False if remote_execution else True,
                        block_type=block_type,
                        kernel_name=kernel_name,
                        pipeline_config=pipeline.get_config_from_yaml(),
                        repo_config=get_repo_config().to_dict(remote=remote_execution),
                        run_tests=run_tests,
                        run_upstream=run_upstream,
                        update_status=False if remote_execution else True,
                        widget=widget,
                    )
                msg_id = client.execute(add_internal_output_info(code))

                WebSocketServer.running_executions_mapping[msg_id] = dict(
                    block_uuid=block_uuid,
                    pipeline_uuid=pipeline_uuid,
                )

                block_output_process_code = get_block_output_process_code(
                    pipeline_uuid,
                    block_uuid,
                    block_type=block_type,
                    kernel_name=kernel_name,
                )
                if block_output_process_code is not None:
                    client.execute(block_output_process_code)

                if run_downstream:
                    for block in block.downstream_blocks:
                        self.on_message(json.dumps(dict(
                            code=block.file.content(),
                            pipeline_uuid=pipeline_uuid,
                            type=block.type,
                            uuid=block.uuid,
                        )))

    @classmethod
    def send_message(self, message: dict) -> None:
        msg_id = message.get('msg_id')
        if msg_id is None:
            return
        msg_id_value = WebSocketServer.running_executions_mapping.get(msg_id, dict())
        block_uuid = msg_id_value.get('block_uuid')
        pipeline_uuid = msg_id_value.get('pipeline_uuid')

        output_dict = dict(uuid=block_uuid, pipeline_uuid=pipeline_uuid)

        message_final = merge_dict(
            message,
            output_dict,
        )

        print(
            f'[{block_uuid}] Sending message for {msg_id} to '
            f'{len(self.clients)} client(s): {message_final}'
        )

        for client in self.clients:
            client.write_message(json.dumps(message_final))



from mage_ai.io.base import BaseSQLConnection, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from pandas import DataFrame
from snowflake.connector import connect
from snowflake.connector.pandas_tools import write_pandas

DEFAULT_LOGIN_TIMEOUT = 20
DEFAULT_NETWORK_TIMEOUT = 20


class Snowflake(BaseSQLConnection):
    """
    Handles data transfer between a Snowflake data warehouse and the Mage app.
    """

    def __init__(self, **kwargs) -> None:
        """
        Initializes settings for connecting to Snowflake data warehouse.
        The following arguments must be provided to the connector, all other
        arguments are optional.

        Required Arguments:
            user (str): Username for the Snowflake user.
            password (str): Login Password for the user.
            account (str): Snowflake account identifier (excluding
            `snowflake-computing.com` suffix).
        """
        if 'login_timeout' not in kwargs:
            kwargs['login_timeout'] = DEFAULT_LOGIN_TIMEOUT
        if 'network_timeout' not in kwargs:
            kwargs['network_timeout'] = DEFAULT_NETWORK_TIMEOUT
        super().__init__(**kwargs)

    def open(self) -> None:
        """
        Opens a connection to Snowflake.
        """
        with self.printer.print_msg('Connecting to Snowflake warehouse'):
            self._ctx = connect(**self.settings)

    def execute(self, query_string: str, **kwargs) -> None:
        """
        Executes any query in the Snowflake data warehouse.

        Args:
            query_string (str): The query to execute on Snowflake's platform.
            **kwargs: Additional parameters to provide to the query
        """
        with self.printer.print_msg(f'Executing query \'{query_string}\''):
            query_string = self._clean_query(query_string)
            with self.conn.cursor() as cur:
                return cur.execute(query_string, **kwargs).fetchall()

    def load(self, query_string: str, limit: int = QUERY_ROW_LIMIT, *args, **kwargs) -> DataFrame:
        """
        Loads data from Snowflake into a Pandas data frame based on the query given.
        This will fail unless a `SELECT` query is provided. This function will load at
        maximum 100,000 rows of data. To operate on more data, consider performing data
        transformations in warehouse.


        Args:
            query_string (str): Query to fetch a table or subset of a table.
            limit (int, Optional): The number of rows to limit the loaded dataframe to. Defaults to 100000.
            *args, **kwargs: Additional parameters to provide to the query

        Returns:
            DataFrame: Data frame associated with the given query.
        """
        with self.printer.print_msg(f'Loading data frame with query \'{query_string}\''):
            query_string = self._clean_query(query_string)
            with self.conn.cursor() as cur:
                return cur.execute(
                    self._enforce_limit(query_string, limit), *args, **kwargs
                ).fetch_pandas_all()

    def export(
        self,
        df: DataFrame,
        table_name: str,
        database: str,
        schema: str,
        if_exists: str = 'append',
        **kwargs,
    ) -> None:
        """
        Exports a Pandas data frame to a Snowflake warehouse based on the table name. If table doesn't
        exist, the table is automatically created.

        Args:
            df (DataFrame): Data frame to export to a Snowflake warehouse.
            table_name (str): Name of the table to export data to (excluding database and schema).
            database (str): Name of the database in which the table is located.
            schema (str): Name of the schema in which the table is located.
            if_exists (str, optional): Specifies export policy if table exists. Either
                - `'fail'`: throw an error.
                - `'replace'`: drops existing table and creates new table of same name.
                - `'append'`: appends data frame to existing table.
            Defaults to `'append'`.
            **kwargs: Additional arguments to pass to writer
        """
        with self.printer.print_msg(
            f'Exporting data frame to table \'{database}.{schema}.{table_name}\''
        ):
            with self._ctx.cursor() as cur:
                cur.execute(f'SHOW TABLES LIKE \'{table_name}\' IN SCHEMA {database}.{schema}')
                if cur.rowcount == 1:
                    if if_exists == 'fail':
                        raise RuntimeError(
                            f'Table {table_name} already exists in the current warehouse, database, schema scenario.'
                        )
                    elif if_exists == 'replace':
                        cur.execute(f'DROP TABLE {table_name}')
                    elif if_exists != 'append':
                        raise ValueError(
                            f'Invalid policy specified for handling existence of table: \'{if_exists}\''
                        )
                elif cur.rowcount > 1:
                    raise ValueError(f'Two or more tables with the name {table_name} are found.')

            auto_create_table = True
            if 'auto_create_table' in kwargs:
                auto_create_table = kwargs.pop(auto_create_table)
                if auto_create_table is None:
                    auto_create_table = True

            write_pandas(
                self.conn,
                df,
                table_name,
                database=database,
                schema=schema,
                auto_create_table=auto_create_table,
                **kwargs,
            )

    @classmethod
    def with_config(cls, config: BaseConfigLoader, **kwargs) -> 'Snowflake':
        """
        Initializes Snowflake client from configuration loader.

        Args:
            config (BaseConfigLoader): Configuration loader object
        """
        return cls(
            user=config[ConfigKey.SNOWFLAKE_USER],
            password=config[ConfigKey.SNOWFLAKE_PASSWORD],
            account=config[ConfigKey.SNOWFLAKE_ACCOUNT],
            warehouse=config[ConfigKey.SNOWFLAKE_DEFAULT_WH],
            database=config[ConfigKey.SNOWFLAKE_DEFAULT_DB],
            schema=config[ConfigKey.SNOWFLAKE_DEFAULT_SCHEMA],
            **kwargs,
        )



from mage_ai.data_cleaner.cleaning_rules.base import BaseRule
from mage_ai.data_cleaner.transformer_actions.constants import (
    ActionType,
    Axis,
)


class RemoveColumnsWithHighEmptyRate(BaseRule):
    default_config = dict(
        missing_rate_threshold=0.8,
    )

    def evaluate(self):
        columns_with_missing_values = []
        columns_with_no_values = []
        for c in self.df_columns:
            if self.statistics.get(f'{c}/count') == 0:
                columns_with_no_values.append(c)
            elif f'{c}/null_value_rate' in self.statistics:
                null_value_rate = self.statistics[f'{c}/null_value_rate']
                if null_value_rate >= self.config('missing_rate_threshold'):
                    columns_with_missing_values.append(c)

        suggestions = []
        if len(columns_with_no_values) > 0:
            suggestions.append(
                self._build_transformer_action_suggestion(
                    'Remove columns with no values',
                    'Remove columns with no values to increase data quality.',
                    ActionType.REMOVE,
                    action_arguments=columns_with_no_values,
                    axis=Axis.COLUMN,
                )
            )
        if len(columns_with_missing_values) > 0:
            suggestions.append(
                self._build_transformer_action_suggestion(
                    'Remove columns with high empty rate',
                    'Remove columns with many missing values may increase data quality.',
                    ActionType.REMOVE,
                    action_arguments=columns_with_missing_values,
                    axis=Axis.COLUMN,
                )
            )
        return suggestions



import asyncio
import os


def main():
    import sys
    command = sys.argv[1]

    if command == 'init':
        from mage_ai.data_preparation.repo_manager import init_repo

        repo_path = os.path.join(os.getcwd(), sys.argv[2])
        init_repo(repo_path)
    elif command == 'start':
        from mage_ai.server.server import main as start_server

        host = None
        port = None
        if len(sys.argv) >= 3:
            repo_path = os.path.join(os.getcwd(), sys.argv[2])
            if len(sys.argv) >= 4:
                host = sys.argv[3]
            if len(sys.argv) >= 5:
                port = sys.argv[4]
        else:
            repo_path = os.getcwd()

        asyncio.run(start_server(
            host=host,
            port=port,
            project=repo_path,
        ))
    elif command == 'run':
        from mage_ai.data_preparation.models.pipeline import Pipeline
        from mage_ai.data_preparation.pipeline_executor import PipelineExecutor

        project_path = sys.argv[2]
        pipeline_uuid = sys.argv[3]
        project_path = os.path.abspath(project_path)
        sys.path.append(os.path.dirname(project_path))
        pipeline = Pipeline(pipeline_uuid, project_path)

        PipelineExecutor.get_executor(pipeline).execute(
            analyze_outputs=False,
            update_status=False,
        )
    elif command == 'test':
        from mage_ai.data_preparation.models.pipeline import Pipeline

        project_path = sys.argv[2]
        pipeline_uuid = sys.argv[3]
        project_path = os.path.abspath(project_path)
        sys.path.append(os.path.dirname(project_path))
        pipeline = Pipeline(pipeline_uuid, project_path)

        asyncio.run(
            pipeline.execute(
                analyze_outputs=False,
                run_tests=True,
                update_status=False,
            )
        )


if __name__ == "__main__":
    main()



from mage_ai.data_cleaner.transformer_actions.udf.base import BaseUDF
import numpy as np

EARTH_RADIUS = 6371


class DistanceBetween(BaseUDF):
    def execute(self):
        def __haversine(lat1, lng1, lat2, lng2):
            lat1, lng1, lat2, lng2 = np.radians([lat1, lng1, lat2, lng2])
            a = np.sin((lat2-lat1)/2.0)**2 + \
                np.cos(lat1) * np.cos(lat2) * np.sin((lng2-lng1)/2.0)**2
            return EARTH_RADIUS * 2 * np.arcsin(np.sqrt(a))
        return __haversine(
            self.df[self.arguments[0]],
            self.df[self.arguments[1]],
            self.df[self.arguments[2]],
            self.df[self.arguments[3]],
        )



from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.variable import Variable, VariableType
from mage_ai.data_preparation.repo_manager import set_repo_path
from mage_ai.data_preparation.variable_manager import (
    VariableManager,
    get_global_variable,
    set_global_variable,
)
from mage_ai.tests.base_test import TestCase
from pandas.util.testing import assert_frame_equal
import os
import pandas as pd
import shutil


class VariableManagerTest(TestCase):
    def setUp(self):
        self.repo_path = os.getcwd() + '/test'
        if not os.path.exists(self.repo_path):
            os.mkdir(self.repo_path)
        return super().setUp()

    def tearDown(self):
        shutil.rmtree(self.repo_path)
        return super().tearDown()

    def test_add_and_get_variable(self):
        self.__create_pipeline('test pipeline 1')
        variable_manager = VariableManager(self.repo_path)
        data1 = {'k1': 'v1', 'k2': 'v2'}
        data2 = pd.DataFrame([
            ['test1', 1],
            ['test2', 2],
        ], columns=['col1', 'col2'])
        data3 = dict(
            metadata=dict(
                column_types=dict(
                    col1='number',
                    col2='text',
                ),
            ),
            statistics=dict(
                count=100,
                count_distinct=50,
            ),
            insights=dict(),
            suggestions=[
                dict(
                    title='Remove outliers',
                )
            ]
        )
        variable_manager.add_variable('test_pipeline_1', 'block1', 'var1', data1)
        variable_manager.add_variable('test_pipeline_1', 'block2', 'var2', data2)
        variable_manager.add_variable(
            'test_pipeline_1',
            'block2',
            'var2',
            data3,
            variable_type=VariableType.DATAFRAME_ANALYSIS,
        )
        self.assertEqual(
            variable_manager.get_variable('test_pipeline_1', 'block1', 'var1'),
            data1,
        )
        assert_frame_equal(
            variable_manager.get_variable('test_pipeline_1', 'block2', 'var2'),
            data2,
        )
        self.assertEqual(
            variable_manager.get_variable(
                'test_pipeline_1',
                'block2',
                'var2',
                variable_type=VariableType.DATAFRAME_ANALYSIS,
            ),
            data3,
        )

    def test_get_variables_by_pipeline(self):
        self.__create_pipeline('test pipeline 2')
        variable_manager = VariableManager(self.repo_path)
        variable_manager.add_variable('test_pipeline_2', 'block1', 'var1', 1)
        variable_manager.add_variable('test_pipeline_2', 'block1', 'var2', 2)
        variable_manager.add_variable('test_pipeline_2', 'block2', 'var3', 3)
        variable_manager.add_variable('test_pipeline_2', 'block2', 'var4', 4)
        self.assertEqual(
            variable_manager.get_variables_by_pipeline('test_pipeline_2'),
            dict(block1=['var1', 'var2'], block2=['var3', 'var4']),
        )

    def test_set_and_get_global_variable(self):
        set_repo_path(self.repo_path)
        self.__create_pipeline('test pipeline 3')
        set_global_variable('test_pipeline_3', 'var1', 1)
        set_global_variable('test_pipeline_3', 'var2', 'test')
        set_global_variable('test_pipeline_3', 'var3', [1, 2, 3])
        set_global_variable('test_pipeline_3', 'var4', dict(k1='v1', k2='v2'))
        self.assertEqual(get_global_variable('test_pipeline_3', 'var1'), 1)
        self.assertEqual(get_global_variable('test_pipeline_3', 'var2'), 'test')
        self.assertEqual(get_global_variable('test_pipeline_3', 'var3'), [1, 2, 3])
        self.assertEqual(get_global_variable('test_pipeline_3', 'var4'), dict(k1='v1', k2='v2'))

    def __create_pipeline(self, name):
        pipeline = Pipeline.create(name, self.repo_path)
        block1 = Block.create('block1', 'data_loader', self.repo_path)
        block2 = Block.create('block2', 'transformer', self.repo_path)
        pipeline.add_block(block1)
        pipeline.add_block(block2)
        return pipeline



from mage_ai.data_preparation.shared.constants import REPO_PATH_ENV_VAR
from mage_ai.data_preparation.templates.utils import copy_template_directory
from typing import Dict
import os
import sys
import yaml


class RepoConfig:
    def __init__(self, repo_path: str = None):
        self.repo_path = repo_path or get_repo_path()
        self.variables_dir = self.repo_path
        try:
            with open(os.path.join(self.repo_path, 'metadata.yaml')) as f:
                repo_config = yaml.full_load(f) or {}
            self.variables_dir = repo_config.get('variables_dir', self.repo_path)
            if self.variables_dir is not None and not self.variables_dir.startswith('s3'):
                self.variables_dir = os.path.abspath(
                    os.path.join(self.repo_path, self.variables_dir),
                )
            self.remote_variables_dir = repo_config.get('remote_variables_dir')
            self.emr_config = repo_config.get('emr_config')
        except Exception:
            pass

    @classmethod
    def from_dict(self, config_dict: Dict) -> 'RepoConfig':
        repo_config = RepoConfig()
        repo_config.emr_config = config_dict.get('emr_config')
        repo_config.repo_path = config_dict.get('repo_path')
        repo_config.variables_dir = config_dict.get('variables_dir')
        repo_config.remote_variables_dir = config_dict.get('remote_variables_dir')
        return repo_config

    def to_dict(self, remote: bool = False) -> Dict:
        return dict(
            emr_config=self.emr_config,
            repo_path=self.repo_path,
            variables_dir=self.remote_variables_dir if remote else self.variables_dir,
            remote_variables_dir=self.remote_variables_dir,
        )


def init_repo(repo_path: str) -> None:
    """
    Initialize a repository under the current path.
    """
    if os.path.exists(repo_path):
        return

    copy_template_directory('repo', repo_path)

    from mage_ai.data_preparation.models.pipeline import Pipeline
    Pipeline.create('default_pipeline', repo_path)


def get_repo_path() -> str:
    return os.getenv(REPO_PATH_ENV_VAR) or os.getcwd()


def get_repo_config(repo_path=None) -> RepoConfig:
    return RepoConfig(repo_path=repo_path)


def set_repo_path(repo_path: str) -> None:
    os.environ[REPO_PATH_ENV_VAR] = repo_path
    sys.path.append(os.path.dirname(repo_path))



{% extends "transformers/transformer_actions/action.jinja" %}
{% block action %}
    """
    Execute Transformer Action: ActionType.COUNT_DISTINCT

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/actions/transformer_actions/README.md#aggregation-actions
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.COUNT_DISTINCT,
        action_code='',  # Enter filtering condition on rows before aggregation
        arguments=[],  # Enter the columns to compute aggregate over
        axis=Axis.COLUMN,
        options={'groupby_columns': []},  # Enter columns to group by
        outputs=[
            # The number of outputs below must match the number of arguments
            {'uuid': 'new_aggregate_column', 'column_type': 'number'},
        ],
    )
{% endblock %}



from mage_ai.data_cleaner.cleaning_rules.base import BaseRule
from mage_ai.data_cleaner.transformer_actions.constants import (
    ActionType,
    Axis,
)


class RemoveColumnsWithSingleValue(BaseRule):

    # Check statistic [feature_uuid]/count_distinct
    def evaluate(self):
        columns_with_single_value = []
        for c in self.df_columns:
            if f'{c}/count_distinct' not in self.statistics:
                continue
            feature_count_distinct = self.statistics[f'{c}/count_distinct']
            if feature_count_distinct == 1:
                columns_with_single_value.append(c)
        suggestions = []
        if len(columns_with_single_value) != 0:
            suggestions.append(
                self._build_transformer_action_suggestion(
                    'Remove columns with single value',
                    'Remove columns with a single unique value to reduce the amount of redundant data.',
                    ActionType.REMOVE,
                    action_arguments=columns_with_single_value,
                    axis=Axis.COLUMN,
                )
            )
        return suggestions



from enum import Enum


class ColumnType(str, Enum):
    CATEGORY = 'category'
    CATEGORY_HIGH_CARDINALITY = 'category_high_cardinality'
    DATETIME = 'datetime'
    EMAIL = 'email'
    LIST = 'list'
    NUMBER = 'number'
    NUMBER_WITH_DECIMALS = 'number_with_decimals'
    PHONE_NUMBER = 'phone_number'
    TEXT = 'text'
    TRUE_OR_FALSE = 'true_or_false'
    ZIP_CODE = 'zip_code'


CATEGORICAL_TYPES = frozenset(
    [ColumnType.CATEGORY, ColumnType.CATEGORY_HIGH_CARDINALITY, ColumnType.TRUE_OR_FALSE]
)
NUMBER_TYPES = frozenset([ColumnType.NUMBER, ColumnType.NUMBER_WITH_DECIMALS])
STRING_TYPES = frozenset(
    [ColumnType.EMAIL, ColumnType.PHONE_NUMBER, ColumnType.TEXT, ColumnType.ZIP_CODE]
)



from mage_ai.data_cleaner.transformer_actions.base import BaseAction
from mage_ai.data_cleaner.transformer_actions.constants import ActionType, Axis
from mage_ai.data_cleaner.transformer_actions.utils import build_transformer_action
from pandas import DataFrame

if 'transformer' not in globals():
    from mage_ai.data_preparation.decorators import transformer


@transformer
def execute_transformer_action(df: DataFrame, *args, **kwargs) -> DataFrame:
    """
    Execute Transformer Action: ActionType.MAX

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/actions/transformer_actions/README.md#aggregation-actions
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.MAX,
        action_code='',  # Enter filtering condition on rows before aggregation
        arguments=[],  # Enter the columns to compute aggregate over
        axis=Axis.COLUMN,
        options={'groupby_columns': []},  # Enter columns to group by
        outputs=[
            # The number of outputs below must match the number of arguments
            {'uuid': 'new_aggregate_column', 'column_type': 'number_with_decimals'},
        ],
    )

    return BaseAction(action).execute(df)



from io import StringIO
from mage_ai.io.base import BaseSQLConnection, ExportWritePolicy, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from mage_ai.io.export_utils import (
    BadConversionError,
    clean_df_for_export,
    gen_table_creation_query,
    infer_dtypes,
    PandasTypes,
)
from pandas import DataFrame, read_sql, Series
from psycopg2 import connect
import numpy as np


class Postgres(BaseSQLConnection):
    """
    Handles data transfer between a PostgreSQL database and the Mage app.
    """

    def __init__(
        self,
        dbname: str,
        user: str,
        password: str,
        host: str,
        port: str = None,
        verbose=True,
        **kwargs,
    ) -> None:
        """
        Initializes the data loader.

        Args:
            dbname (str): The name of the database to connect to.
            user (str): The user with which to connect to the database with.
            password (str): The login password for the user.
            host (str): Path to host address for database.
            port (str): Port on which the database is running.
            **kwargs: Additional settings for creating SQLAlchemy engine and connection
        """
        super().__init__(
            verbose=verbose,
            dbname=dbname,
            user=user,
            password=password,
            host=host,
            port=port,
            **kwargs,
        )

    def open(self) -> None:
        """
        Opens a connection to the PostgreSQL database specified by the parameters.
        """
        with self.printer.print_msg('Opening connection to PostgreSQL database'):
            self._ctx = connect(**self.settings)

    def execute(self, query_string: str, **query_vars) -> None:
        """
        Sends query to the connected database.

        Args:
            query_string (str): SQL query string to apply on the connected database.
            query_vars: Variable values to fill in when using format strings in query.
        """
        with self.printer.print_msg(f'Executing query \'{query_string}\''):
            query_string = self._clean_query(query_string)
            with self.conn.cursor() as cur:
                cur.execute(query_string, **query_vars)

    def load(self, query_string: str, limit: int = QUERY_ROW_LIMIT, **kwargs) -> DataFrame:
        """
        Loads data from the connected database into a Pandas data frame based on the query given.
        This will fail if the query returns no data from the database. This function will load at
        maximum 100,000 rows of data. To operate on more data, consider performing data
        transformations in warehouse.

        Args:
            query_string (str): Query to execute on the database.
            limit (int, Optional): The number of rows to limit the loaded dataframe to. Defaults to 100000.
            **kwargs: Additional query parameters.

        Returns:
            DataFrame: The data frame corresponding to the data returned by the given query.
        """
        with self.printer.print_msg(f'Loading data frame with query \'{query_string}\''):
            query_string = self._clean_query(query_string)
            return read_sql(self._enforce_limit(query_string, limit), self.conn, **kwargs)

    def export(
        self,
        df: DataFrame,
        schema_name: str,
        table_name: str,
        if_exists: ExportWritePolicy = ExportWritePolicy.REPLACE,
        index: bool = False,
    ) -> None:
        """
        Exports dataframe to the connected database from a Pandas data frame. If table doesn't
        exist, the table is automatically created. If the schema doesn't exist, the schema is also created.

        Args:
            schema_name (str): Name of the schema of the table to export data to.
            table_name (str): Name of the table to insert rows from this data frame into.
            if_exists (ExportWritePolicy): Specifies export policy if table exists. Either
                - `'fail'`: throw an error.
                - `'replace'`: drops existing table and creates new table of same name.
                - `'append'`: appends data frame to existing table. In this case the schema must match the original table.
            Defaults to `'replace'`.
            index (bool): If true, the data frame index is also exported alongside the table. Defaults to False.
            **kwargs: Additional query parameters.
        """
        if index:
            df = df.reset_index()

        dtypes = infer_dtypes(df)
        df = clean_df_for_export(df, self.clean, dtypes)

        with self.printer.print_msg(
            f'Exporting data frame to table \'{schema_name}.{table_name}\''
        ):
            buffer = StringIO()
            table_exists = self.__table_exists(schema_name, table_name)

            with self.conn.cursor() as cur:
                cur.execute(f'CREATE SCHEMA IF NOT EXISTS {schema_name};')
                if table_exists:
                    if if_exists == ExportWritePolicy.FAIL:
                        raise ValueError(
                            f'Table \'{schema_name}.{table_name}\' already exists in database'
                        )
                    elif if_exists == ExportWritePolicy.REPLACE:
                        cur.execute(f'DELETE FROM {schema_name}.{table_name}')
                else:
                    db_dtypes = {col: self.get_type(df[col], dtypes[col]) for col in dtypes}
                    query = gen_table_creation_query(db_dtypes, schema_name, table_name)
                    cur.execute(query)

                df.to_csv(buffer, index=False, header=False)
                buffer.seek(0)
                cur.copy_expert(
                    f'COPY {schema_name}.{table_name} FROM STDIN (FORMAT csv, DELIMITER \',\', NULL \'\');',
                    buffer,
                )
            self.conn.commit()

    def __table_exists(self, schema_name: str, table_name: str) -> bool:
        """
        Returns whether the specified table exists.

        Args:
            schema_name (str): Name of the schema the table belongs to.
            table_name (str): Name of the table to check existence of.

        Returns:
            bool: True if the table exists, else False.
        """
        with self.conn.cursor() as cur:
            cur.execute(
                f'SELECT * FROM pg_tables WHERE schemaname = \'{schema_name}\' AND tablename = \'{table_name}\''
            )
            return bool(cur.rowcount)

    def clean(self, column: Series, dtype: str) -> Series:
        """
        Cleans column in order to write data frame to PostgreSQL database

        Args:
            column (Series): Column to clean
            dtype (str): The pandas data types of this column

        Returns:
            Series: Cleaned column
        """
        if dtype == PandasTypes.CATEGORICAL:
            return column.astype(str)
        elif dtype in (PandasTypes.TIMEDELTA, PandasTypes.TIMEDELTA64, PandasTypes.PERIOD):
            return column.view(int)
        else:
            return column

    def get_type(self, column: Series, dtype: str) -> str:
        """
        Maps pandas Data Frame column to PostgreSQL type

        Args:
            series (Series): Column to map
            dtype (str): Pandas data type of this column

        Raises:
            ConversionError: Returned if this type cannot be converted to a PostgreSQL data type

        Returns:
            str: PostgreSQL data type for this column
        """
        if dtype in (
            PandasTypes.MIXED,
            PandasTypes.UNKNOWN_ARRAY,
            PandasTypes.COMPLEX,
        ):
            raise BadConversionError(
                f'Cannot convert column \'{column.name}\' with data type \'{dtype}\' to a PostgreSQL datatype.'
            )
        elif dtype in (PandasTypes.DATETIME, PandasTypes.DATETIME64):
            try:
                if column.dt.tz:
                    return 'timestamptz'
            except AttributeError:
                pass
            return 'timestamp'
        elif dtype == PandasTypes.TIME:
            try:
                if column.dt.tz:
                    return 'timetz'
            except AttributeError:
                pass
            return 'time'
        elif dtype == PandasTypes.DATE:
            return 'date'
        elif dtype == PandasTypes.STRING:
            return 'text'
        elif dtype == PandasTypes.CATEGORICAL:
            return 'text'
        elif dtype == PandasTypes.BYTES:
            return 'bytea'
        elif dtype in (PandasTypes.FLOATING, PandasTypes.DECIMAL, PandasTypes.MIXED_INTEGER_FLOAT):
            return 'double precision'
        elif dtype == PandasTypes.INTEGER:
            max_int, min_int = column.max(), column.min()
            if np.int16(max_int) == max_int and np.int16(min_int) == min_int:
                return 'smallint'
            elif np.int32(max_int) == max_int and np.int32(min_int) == min_int:
                return 'integer'
            else:
                return 'bigint'
        elif dtype == PandasTypes.BOOLEAN:
            return 'boolean'
        elif dtype in (PandasTypes.TIMEDELTA, PandasTypes.TIMEDELTA64, PandasTypes.PERIOD):
            return 'bigint'
        else:
            raise ValueError(f'Invalid datatype provided: {dtype}')

    @classmethod
    def with_config(cls, config: BaseConfigLoader) -> 'Postgres':
        """
        Initializes PostgreSQL loader from configuration loader

        Args:
            config (BaseConfigLoader): Configuration loader object
        """
        return cls(
            dbname=config[ConfigKey.POSTGRES_DBNAME],
            user=config[ConfigKey.POSTGRES_USER],
            password=config[ConfigKey.POSTGRES_PASSWORD],
            host=config[ConfigKey.POSTGRES_HOST],
            port=config[ConfigKey.POSTGRES_PORT],
        )



import json
import simplejson
import tornado.web
import traceback


class BaseHandler(tornado.web.RequestHandler):
    model_class = None

    def check_origin(self, origin):
        return True

    def get_bool_argument(self, name, default_value=None):
        value = self.get_argument(name, default_value)
        if type(value) is not str:
            return value
        return value.lower() in ('yes', 'true', 't', '1')

    def options(self, **kwargs):
        self.set_status(204)
        self.finish()

    def set_default_headers(self):
        methods = 'DELETE, GET, PATCH, POST, PUT, OPTIONS'
        self.set_header('Access-Control-Allow-Headers', '*')
        self.set_header('Access-Control-Allow-Methods', methods)
        self.set_header('Access-Control-Allow-Origin', '*')
        self.set_header('Content-Type', 'application/json')

    def write(self, chunk):
        if type(chunk) is dict:
            chunk = simplejson.dumps(chunk, ignore_nan=True)
        super().write(chunk)

    def write_error(self, status_code, **kwargs):
        if status_code == 500:
            self.set_status(200)
            exception = kwargs['exc_info'][1]
            self.write(
                dict(
                    error=dict(
                        code=status_code,
                        errors=traceback.format_stack(),
                        exception=str(exception),
                        message=traceback.format_exc(),
                    ),
                    url_parameters=self.path_kwargs,
                )
            )

    def get_payload(self):
        key = ''
        if self.model_class:
            key = self.model_class.__name__.lower()

        return json.loads(self.request.body).get(key, {})



from .base import BaseHandler
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.models.widget import Widget
from mage_ai.data_preparation.repo_manager import get_repo_path


class ApiPipelineWidgetDetailHandler(BaseHandler):
    model_class = Widget

    def put(self, pipeline_uuid, block_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        payload = self.get_payload()

        widget = pipeline.get_block(block_uuid, widget=True)
        if widget is None:
            raise Exception(f'Widget {block_uuid} does not exist in pipeline {pipeline_uuid}')

        widget.update(payload)
        if payload.get('configuration'):
            widget.configuration = payload['configuration']
            pipeline.save()

        self.write(dict(widget=widget.to_dict(include_content=True)))

    def delete(self, pipeline_uuid, block_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        widget = pipeline.get_block(block_uuid, widget=True)
        if widget is None:
            raise Exception(f'widget {block_uuid} does not exist in pipeline {pipeline_uuid}')
        widget.delete()
        self.write(dict(widget=widget.to_dict()))


class ApiPipelineWidgetListHandler(BaseHandler):
    model_class = Widget

    def get(self, pipeline_uuid):
        include_outputs = self.get_bool_argument('include_outputs', True)

        pipeline = Pipeline.get(pipeline_uuid)
        collection = [widget.to_dict(
            include_content=True,
            include_outputs=include_outputs,
        ) for widget in pipeline.widgets_by_uuid.values()]

        self.write(dict(widgets=collection))
        self.finish()

    def post(self, pipeline_uuid):
        pipeline = Pipeline.get(pipeline_uuid)

        payload = self.get_payload()

        resource = Widget.create(
            payload.get('name') or payload.get('uuid'),
            payload.get('type'),
            get_repo_path(),
            config=payload.get('config'),
            pipeline=pipeline,
            priority=payload.get('priority'),
            upstream_block_uuids=payload.get('upstream_blocks', []),
        )

        if payload.get('content'):
            resource.update_content(payload['content'], widget=True)

        if payload.get('configuration'):
            resource.configuration = payload['configuration']
            pipeline.save()

        self.write(dict(widget=resource.to_dict(include_content=True)))



import boto3
import boto3.s3.transfer as s3transfer
import botocore

MAX_POOL_CONNECTIONS = 100
MAX_KEYS = 10000


class Client():
    def __init__(self, bucket, **kwargs):
        self.bucket = bucket
        self.client = kwargs.get(
            'client',
            boto3.client(
                's3', 
                config=botocore.client.Config(max_pool_connections=MAX_POOL_CONNECTIONS),
            ),
        )
        self.transfer_config = s3transfer.TransferConfig(
            use_threads=True,
            max_concurrency=MAX_POOL_CONNECTIONS,
        )

    def download_file(self, object_key, filename_destination):
        return self.resource().Bucket(self.bucket).download_file(object_key, filename_destination)

    def resource(self):
        return boto3.resource('s3')

    def read(self, object_key):
        return self.get_object(object_key).read()

    def get_object(self, object_key):
        return self.client.get_object(Bucket=self.bucket, Key=object_key)['Body']

    def delete_objects(self, prefix):
        keys = self.list_objects(prefix)
        self.client.delete_objects(
            Bucket=self.bucket,
            Delete={
                'Objects': [{'Key': key} for key in keys],
            },
        )

    def list_folders(self, prefix, delimiter='/'):
        keys = []
        response = self.client.list_objects_v2(
            Bucket=self.bucket,
            MaxKeys=MAX_KEYS,
            Prefix=prefix,
            Delimiter=delimiter,
        )
        if response.get('CommonPrefixes'):
            for obj in response['CommonPrefixes']:
                keys.append(obj['Prefix'])
        return keys

    def list_objects(self, prefix):
        keys = []
        response = self.client.list_objects_v2(Bucket=self.bucket, MaxKeys=MAX_KEYS, Prefix=prefix)
        if response.get('Contents'):
            for obj in response['Contents']:
                keys.append(obj['Key'])
        return keys

    def upload(self, object_key, content):
        return self.client.put_object(
            Body=content,
            Bucket=self.bucket,
            Key=object_key,
        )

    def upload_object(self, object_key, file):
        return self.client.upload_fileobj(file, Bucket=self.bucket, Key=object_key)



from mage_ai.data_cleaner.column_types.constants import NUMBER_TYPES, ColumnType
from mage_ai.data_cleaner.transformer_actions.constants import CURRENCY_SYMBOLS
from mage_ai.shared.custom_types import FrozenDict
from pandas.core.indexes.frozen import FrozenList
from typing import Any, List, Union
import pandas as pd
import numpy as np
import re

COLUMN_NAME_QUOTE_CHARS = '+=-*&^%$! ?~|<>(){}[],.'
LIST_SPLIT = re.compile(r'\s*,\s*')
LIST_TYPES = frozenset([list, set, tuple])
NONE_TYPES = frozenset([None, np.nan])
PAREN_LIKE_OPEN = frozenset(['[', '('])
PAREN_LIKE_CLOSE = frozenset([']', ')'])


def clean_series(series, column_type, dropna=True):
    if series.dtype == 'object':
        series_cleaned = series.apply(lambda x: x.strip(' \'\"') if type(x) is str else x)
        series_cleaned = series_cleaned.map(
            lambda x: x if (not isinstance(x, str) or x != '') else np.nan
        )
    else:
        series_cleaned = series
    if dropna:
        series_cleaned = series_cleaned.dropna()

    if series_cleaned.count() == 0:
        return series_cleaned

    dtype = type(series_cleaned.dropna().iloc[0])
    if column_type == ColumnType.NUMBER or column_type == ColumnType.NUMBER_WITH_DECIMALS:
        is_percent = False
        if dtype is str:
            series_cleaned = series_cleaned.str.replace(',', '')
            if series_cleaned.str.count(CURRENCY_SYMBOLS).sum() != 0:
                series_cleaned = series_cleaned.str.replace(CURRENCY_SYMBOLS, '')
            elif series_cleaned.str.contains('%').sum() != 0:
                is_percent = True
                series_cleaned = series_cleaned.str.replace('%', '')
            series_cleaned = series_cleaned.str.replace(' ', '')
        if column_type == ColumnType.NUMBER:
            try:
                series_cleaned = series_cleaned.astype(int)
            except ValueError:
                series_cleaned = series_cleaned.astype(float)
        else:
            series_cleaned = series_cleaned.astype(float)
        if is_percent:
            series_cleaned /= 100
    elif column_type == ColumnType.DATETIME:
        series_cleaned = pd.to_datetime(series_cleaned, errors='coerce', infer_datetime_format=True)
    elif column_type == ColumnType.PHONE_NUMBER and dtype is not str:
        series_cleaned = series_cleaned.astype(str)
        series_cleaned = series_cleaned.str.replace(r'\.\d*', '', regex=True)
    elif column_type == ColumnType.ZIP_CODE and dtype is not str:
        series_cleaned = series_cleaned.astype(str)
    elif column_type == ColumnType.LIST:
        series_cleaned = series_cleaned.apply(parse_list)
    return series_cleaned


def clean_dataframe(df, column_types, dropna=True):
    return df.apply(lambda col: clean_series(col, column_types[col.name], dropna=dropna))


def is_dataframe(df):
    return type(df) is pd.DataFrame or is_spark_dataframe(df)


def is_numeric_dtype(df, column, column_type):
    return column_type in NUMBER_TYPES or issubclass(df[column].dtype.type, np.number)


def is_spark_dataframe(df):
    return type(df).__module__ == 'pyspark.sql.dataframe'


def __parse_element(element: str) -> Any:
    if element == 'nan' or element == 'np.nan':
        return np.nan
    else:
        try:
            return __resolve_type(eval(element))
        except:
            return None


def __resolve_type(element: Any) -> Any:
    if isinstance(element, dict):
        return FrozenDict(element)
    return element


def parse_list(list_literal: Union[str, List[Any]]) -> FrozenList:
    dtype = type(list_literal)
    if dtype is FrozenList:
        return list_literal
    elif dtype in LIST_TYPES:
        return FrozenList([__resolve_type(literal) for literal in list_literal])
    elif list_literal in NONE_TYPES:
        return list_literal
    elif dtype is not str:
        return FrozenList([list_literal])
    if list_literal[0] not in PAREN_LIKE_OPEN or list_literal[-1] not in PAREN_LIKE_CLOSE:
        return list_literal
    list_literal = list_literal.strip('[]() ')
    if list_literal == '':
        return FrozenList([])
    return FrozenList([__parse_element(element) for element in LIST_SPLIT.split(list_literal)])


def wrap_column_name(name: str) -> str:
    if any(symbol in name for symbol in COLUMN_NAME_QUOTE_CHARS):
        name = f'"{name}"'
    return name



import random


def batch(iterable, n=1):
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx:min(ndx + n, l)]


def difference(li1, li2):
    li1_lookup = set(li1)
    li2_lookup = set(li2)
    return [i for i in li1 + li2 if i not in li1_lookup or i not in li2_lookup]


def flatten(arr):
    return [item for sublist in arr for item in sublist]


def find(condition, arr, map=None):
    try:
        return next(map(x) if map else x for x in arr if condition(x))
    except StopIteration:
        return None


def sample(arr):
    return arr[random.randrange(0, len(arr))]


def subtract(arr1, arr2):
    arr2_lookup = set(arr2)
    return [i for i in arr1 if i not in arr2_lookup]



{% extends "data_loaders/default.jinja" %}
{% block imports %}
import io
import pandas as pd
import requests
{{ super() -}}
{% endblock %}

{% block content %}
@data_loader
def load_data_from_api() -> DataFrame:
    """
    Template for loading data from API
    """
    url = ''

    response = requests.get(url)
    return pd.read_csv(io.StringIO(response.text), sep=',')
{% endblock %}



from mage_ai.server.constants import SERVER_HOST, SERVER_PORT
import logging
import os
import sys

MAX_NUM_OF_ROWS = 100_000

logger = logging.getLogger(__name__)


def launch(
    host=SERVER_HOST,
    port=SERVER_PORT,
    inline=True,
    api_key=None,
    notebook_type=None,
    iframe_host=None,
    iframe_port=None,
    config={},
):
    from mage_ai.server.app import launch as launch_flask
    from mage_ai.server.utils.frontend_renderer import (
        NotebookType,
        display_inline_iframe,
        infer_notebook_type,
        update_frontend_urls,
    )
    iframe_host = iframe_host or host
    iframe_port = iframe_port or port
    if notebook_type is None:
        notebook_type = infer_notebook_type()
    if notebook_type == NotebookType.DATABRICKS:
        host = '0.0.0.0'
        update_frontend_urls(
            host=iframe_host,
            port=iframe_port,
            notebook_type=notebook_type,
            config=config,
        )
    elif notebook_type == NotebookType.SAGEMAKER:
        update_frontend_urls(
            host=host,
            port=port,
            notebook_type=notebook_type,
            config=config,
        )
    thread = launch_flask(mage_api_key=api_key, host=host, port=port)
    if inline:
        display_inline_iframe(
            host=iframe_host,
            port=iframe_port,
            notebook_type=notebook_type,
            config=config,
        )

    return thread


def remote_sync(api_key=None):
    from mage_ai.server.app import sync_pipelines
    sync_pipelines(api_key)


def kill():
    from mage_ai.server.app import kill as kill_flask
    kill_flask()


def connect_data(df, name, verbose=False):
    from mage_ai.data_cleaner.shared.utils import is_spark_dataframe
    from mage_ai.server.app import connect_df

    if is_spark_dataframe(df):
        # Convert pyspark dataframe to pandas
        df_spark = df
        row_count = df_spark.count()
        if row_count >= MAX_NUM_OF_ROWS:
            sample_fraction = MAX_NUM_OF_ROWS / row_count
            df = df_spark.sample(withReplacement=False, fraction=sample_fraction).toPandas()
        else:
            df = df_spark.toPandas()

    if df.shape[0] > MAX_NUM_OF_ROWS:
        feature_set, _ = connect_df(
            df.sample(MAX_NUM_OF_ROWS).reset_index(drop=True),
            name,
            verbose=verbose,
        )
    else:
        feature_set, _ = connect_df(df, name, verbose=verbose)
    return feature_set


def clean(
    df,
    name=None,
    pipeline_uuid=None,
    pipeline_path=None,
    remote_pipeline_uuid=None,
    api_key=None,
    verbose=False,
):
    from mage_ai.server.app import (
        clean_df,
        clean_df_with_pipeline,
    )
    if pipeline_uuid is not None:
        df_clean = clean_df_with_pipeline(df, id=pipeline_uuid, verbose=verbose)
    elif pipeline_path is not None:
        df_clean = clean_df_with_pipeline(df, path=pipeline_path, verbose=verbose)
    elif remote_pipeline_uuid is not None:
        df_clean = clean_df_with_pipeline(
            df, remote_id=remote_pipeline_uuid, mage_api_key=api_key, verbose=verbose
        )
    else:
        _, df_clean = clean_df(df, name=name, verbose=verbose)
    return df_clean


# --------------- Data preparation methods --------------- #


def run(pipeline_uuid: str, project_path: str = None, **global_vars) -> None:
    from mage_ai.data_preparation.models.pipeline import Pipeline
    from mage_ai.data_preparation.pipeline_executor import PipelineExecutor

    project_path = os.getcwd() if project_path is None else os.path.abspath(project_path)
    sys.path.append(os.path.dirname(project_path))
    pipeline = Pipeline(pipeline_uuid, project_path)
    PipelineExecutor.get_executor(pipeline).execute(
        analyze_outputs=False,
        global_vars=global_vars,
        update_status=False,
    )



from mage_ai.data_cleaner.cleaning_rules.remove_columns_with_high_empty_rate import (
    RemoveColumnsWithHighEmptyRate,
)
from mage_ai.tests.base_test import TestCase
import numpy as np
import pandas as pd


class RemoveColumnWithHighEmptyRateTests(TestCase):
    def test_evaluate(self):
        result = RemoveColumnsWithHighEmptyRate(
            **self.__create_test_data(),
        ).evaluate()

        self.assertEqual(
            result,
            [
                dict(
                    title='Remove columns with high empty rate',
                    message='Remove columns with many missing values may increase data quality.',
                    status='not_applied',
                    action_payload=dict(
                        action_type='remove',
                        action_arguments=['deleted_at'],
                        action_code='',
                        action_options={},
                        action_variables={},
                        axis='column',
                        outputs=[],
                    ),
                )
            ],
        )

    def test_evaluate_with_custom_config(self):
        test_data = self.__create_test_data()
        result = RemoveColumnsWithHighEmptyRate(
            test_data['df'],
            test_data['column_types'],
            test_data['statistics'],
            custom_config=dict(missing_rate_threshold=0.6),
        ).evaluate()

        self.assertEqual(
            result,
            [
                dict(
                    title='Remove columns with high empty rate',
                    message='Remove columns with many missing values may increase data quality.',
                    status='not_applied',
                    action_payload=dict(
                        action_type='remove',
                        action_arguments=['country', 'deleted_at'],
                        action_code='',
                        action_options={},
                        action_variables={},
                        axis='column',
                        outputs=[],
                    ),
                )
            ],
        )

    def __create_test_data(self):
        df = pd.DataFrame(
            [
                [1, np.NaN, '2022-01-01'],
                [2, 'US', np.NaN],
                [3, 'US', np.NaN],
                [4, np.NaN, np.NaN],
                [5, np.NaN, np.NaN],
            ],
            columns=['id', 'country', 'deleted_at'],
        )
        column_types = {
            'id': 'number',
            'country': 'category',
            'deleted_at': 'datetime',
        }
        statistics = {
            'id/null_value_rate': 0,
            'country/null_value_rate': 0.6,
            'deleted_at/null_value_rate': 0.8,
        }
        return dict(
            df=df,
            column_types=column_types,
            statistics=statistics,
        )



from mage_ai.data_cleaner.transformer_actions.helpers import extract_join_feature_set_version_id
from mage_ai.tests.base_test import TestCase


class ColumnTests(TestCase):
    def test_extract_join_feature_set_version_id(self):
        payload1 = dict(
            action_type='join',
            action_arguments=[100],
            action_options=dict(
                left_on=['user_id'],
                right_on=['id'],
            ),
        )
        payload2 = dict(
            action_type='join',
            action_arguments=['%{1}'],
            action_options=dict(
                left_on=['user_id'],
                right_on=['id'],
            ),
            action_variables={
                '1': {
                    'id': 200,
                    'type': 'feature_set_version',
                },
            },
        )
        payload3 = dict(
            action_type='filter',
        )
        fsv_id1 = extract_join_feature_set_version_id(payload1)
        fsv_id2 = extract_join_feature_set_version_id(payload2)
        fsv_id3 = extract_join_feature_set_version_id(payload3)
        self.assertEqual(fsv_id1, 100)
        self.assertEqual(fsv_id2, 200)
        self.assertEqual(fsv_id3, None)



import jinja2
import os
import shutil


template_env = jinja2.Environment(
    loader=jinja2.FileSystemLoader(os.path.dirname(__file__)),
    lstrip_blocks=True,
    trim_blocks=True,
)


def copy_template_directory(template_path: str, dest_path: str) -> None:
    """
    Copies a template directory structure from source to destination.

    Args:
        template_path (str): Source directory for the template to copy.
        dest_path (str): Destination directory to copy template to.

    Raises:
        IOError: Raises IOError if template could not be found.
    """
    template_path = os.path.join(
        os.path.dirname(__file__),
        template_path,
    )
    if not os.path.exists(template_path):
        raise IOError(f'Could not find templates for {template_path}.')
    shutil.copytree(template_path, dest_path)


def read_template_file(template_path: str) -> jinja2.Template:
    """
    Reads template source code into a string

    Args:
        template_path (str): File path of template to load relative to `templates` package

    Returns:
        jinja2.Template: Template source object
    """
    return template_env.get_template(template_path)


def write_template(template_source: str, dest_path: str) -> None:
    """
    Writes template source code to destination file

    Args:
        template_source (str): Template source code to write to file
        dest_path (str): Destination file to write template source code to.
    """
    with open(dest_path, 'w') as foutput:
        foutput.write(template_source)




from mage_ai.data_cleaner.transformer_actions.base import BaseAction
from mage_ai.data_cleaner.transformer_actions.row import (
    drop_duplicates,
    # explode,
    filter_rows,
    sort_rows,
    remove_row,
)
from mage_ai.tests.base_test import TestCase
from pandas.util.testing import assert_frame_equal
import numpy as np
import pandas as pd


class RowTests(TestCase):
    def test_drop_duplicates(self):
        df = pd.DataFrame(
            [
                [0, False, 'a'],
                [1, True, 'b'],
                [1, True, 'c'],
                [0, True, 'd'],
                [1, True, 'b'],
            ],
            columns=[
                'integer',
                'boolean',
                'string',
            ],
        )

        test_cases = [
            (dict(action_arguments=['integer']), df.iloc[[3, 4]]),
            (
                dict(action_arguments=['integer'], action_options=dict(keep='first')),
                df.iloc[[0, 1]],
            ),
            (dict(action_arguments=['boolean']), df.iloc[[0, 4]]),
            (
                dict(action_arguments=['boolean'], action_options=dict(keep='first')),
                df.iloc[[0, 1]],
            ),
            (dict(action_arguments=['integer', 'boolean']), df.iloc[[0, 3, 4]]),
            (dict(action_arguments=[]), df.iloc[[0, 2, 3, 4]]),
        ]

        for action, val in test_cases:
            self.assertTrue(drop_duplicates(df, action).equals(val))

    # def test_explode(self):
    #     df = pd.DataFrame([
    #         ['(a, b, c)'],
    #         ['[b, c, d]'],
    #         [' e, f '],
    #     ], columns=['tags'])
    #     action = dict(
    #         action_arguments=['tags'],
    #         action_options={
    #             'separator': ',',
    #         },
    #         outputs=[
    #             dict(
    #                 uuid='tag',
    #                 column_type='text',
    #             ),
    #         ],
    #     )
    #     df_new = explode(df, action)
    #     df_expected = pd.DataFrame([
    #         ['a', '(a, b, c)'],
    #         ['b', '(a, b, c)'],
    #         ['c', '(a, b, c)'],
    #         ['b', '[b, c, d]'],
    #         ['c', '[b, c, d]'],
    #         ['d', '[b, c, d]'],
    #         ['e', ' e, f '],
    #         ['f', ' e, f '],
    #     ], columns=['tag', 'tags'])
    #     assert_frame_equal(df_new.reset_index(drop=True), df_expected)

    def test_filter_rows(self):
        df = pd.DataFrame(
            [
                [0, False, 'a'],
                [1, True, 'b'],
            ],
            columns=[
                'integer',
                'boolean',
                'string',
            ],
        )

        test_cases = [
            ([0, False, 'a'], 'integer == 0'),
            ([0, False, 'a'], 'string == \'a\''),
            ([1, True, 'b'], 'boolean == True'),
            ([1, True, 'b'], 'integer >= 1'),
            ([1, True, 'b'], 'integer >= 1 and boolean == True'),
            ([1, True, 'b'], 'integer >= 1 and (boolean == False or string == \'b\')'),
        ]

        for val, query in test_cases:
            self.assertEqual(
                val,
                filter_rows(df, dict(action_code=query)).iloc[0].values.tolist(),
            )

    def test_filter_rows_is_null(self):
        df = pd.DataFrame(
            [
                [None, False, 'a'],
                [2, True, 'b'],
                [3, False, 'c'],
                [1, None, 'a'],
                [2, True, 'b'],
                [3, '', 'c'],
                [1, False, None],
                [2, True, 'b'],
                [3, False, ''],
            ],
            columns=[
                'integer',
                'boolean',
                'string',
            ],
        )

        integer_rows = filter_rows(
            df,
            dict(action_code='integer == null'),
            original_df=df,
        ).values.tolist()
        self.assertEqual(len(integer_rows), 1)
        self.assertEqual(integer_rows[0][1], False)
        self.assertEqual(integer_rows[0][2], 'a')

        boolean_rows = filter_rows(
            df,
            dict(action_code='boolean == null'),
            original_df=df,
        ).values.tolist()
        self.assertEqual(len(boolean_rows), 2)
        self.assertEqual(boolean_rows[0][0], 1.0)
        self.assertEqual(boolean_rows[0][1], None)
        self.assertEqual(boolean_rows[0][2], 'a')
        self.assertEqual(boolean_rows[1][0], 3.0)
        self.assertEqual(boolean_rows[1][1], '')
        self.assertEqual(boolean_rows[1][2], 'c')

        string_rows = filter_rows(
            df,
            dict(action_code='string == null'),
            original_df=df,
        ).values.tolist()
        self.assertEqual(len(string_rows), 2)
        self.assertEqual(string_rows[0][0], 1.0)
        self.assertEqual(string_rows[0][1], False)
        self.assertEqual(string_rows[0][2], None)
        self.assertEqual(string_rows[1][0], 3.0)
        self.assertEqual(string_rows[1][1], False)
        self.assertEqual(string_rows[1][2], '')

    def test_filter_rows_is_not_null(self):
        df = pd.DataFrame(
            [
                [None, False, 'a'],
                [2, True, 'b'],
                [3, False, 'c'],
                [1, None, 'a'],
                [2, True, 'b'],
                [3, '', 'c'],
                [1, False, None],
                [2, True, 'b'],
                [3, False, ''],
            ],
            columns=[
                'integer',
                'boolean',
                'string',
            ],
        )
        integer_rows = filter_rows(
            df,
            dict(action_code='integer != null'),
            original_df=df,
        )['integer'].values.tolist()
        self.assertEqual(
            integer_rows,
            [
                2,
                3,
                1,
                2,
                3,
                1,
                2,
                3,
            ],
        )

        boolean_rows = filter_rows(
            df,
            dict(action_code='boolean != null'),
            original_df=df,
        )['boolean'].values.tolist()
        self.assertEqual(
            boolean_rows,
            [
                False,
                True,
                False,
                True,
                False,
                True,
                False,
            ],
        )

        string_rows = filter_rows(
            df,
            dict(action_code='string != null'),
            original_df=df,
        )['string'].values.tolist()
        self.assertEqual(
            string_rows,
            [
                'a',
                'b',
                'c',
                'a',
                'b',
                'c',
                'b',
            ],
        )

    def test_filter_row_contains_string(self):
        df = pd.DataFrame(
            [
                ['fsdijfosidjfiosfj'],
                ['abc@123.com'],
                [np.NaN],
                ['fsdfsdfdsfdsf'],
                ['xyz@456.com'],
            ],
            columns=[
                'id',
            ],
        )
        action = dict(
            action_code='id contains @',
        )
        action2 = dict(
            action_code='id contains \'@\'',
        )
        df_new = filter_rows(df, action, original_df=df).reset_index(drop=True)
        df_new2 = filter_rows(df, action2, original_df=df).reset_index(drop=True)
        df_expected = pd.DataFrame(
            [
                ['abc@123.com'],
                ['xyz@456.com'],
            ],
            columns=[
                'id',
            ],
        )
        assert_frame_equal(df_new, df_expected)
        assert_frame_equal(df_new2, df_expected)

    def test_filter_row_not_contains_string(self):
        df = pd.DataFrame(
            [
                [np.NaN, False],
                ['sfc@mailnet.com', True],
                ['fdss@emailserver.net', True],
                ['fsdfsdfdsfdsf', False],
                ['xyz@mailnet.com', False],
                ['eeeeasdf', True],
            ],
            columns=['email', 'subscription'],
        )
        action = dict(
            action_code='email not contains mailnet',
        )
        action2 = dict(
            action_code='email not contains \'mailnet\'',
        )
        action3 = dict(
            action_code='email not contains @',
        )
        action4 = dict(
            action_code='email not contains \'^e+\w\'',
        )
        action_invalid = dict(action_code='subscription not contains False')
        df_new = filter_rows(df, action, original_df=df).reset_index(drop=True)
        df_new2 = filter_rows(df, action2, original_df=df).reset_index(drop=True)
        df_new3 = filter_rows(df, action3, original_df=df).reset_index(drop=True)
        df_new4 = filter_rows(df, action4, original_df=df).reset_index(drop=True)
        df_expected1 = pd.DataFrame(
            [
                [np.NaN, False],
                ['fdss@emailserver.net', True],
                ['fsdfsdfdsfdsf', False],
                ['eeeeasdf', True],
            ],
            columns=['email', 'subscription'],
        )
        df_expected2 = pd.DataFrame(
            [[np.NaN, False], ['fsdfsdfdsfdsf', False], ['eeeeasdf', True]],
            columns=['email', 'subscription'],
        )
        df_expected3 = pd.DataFrame(
            [
                [np.NaN, False],
                ['sfc@mailnet.com', True],
                ['fdss@emailserver.net', True],
                ['fsdfsdfdsfdsf', False],
                ['xyz@mailnet.com', False],
            ],
            columns=['email', 'subscription'],
        )
        assert_frame_equal(df_new, df_expected1)
        assert_frame_equal(df_new2, df_expected1)
        assert_frame_equal(df_new3, df_expected2)
        assert_frame_equal(df_new4, df_expected3)

        with self.assertRaises(Exception):
            _ = filter_rows(df, action_invalid, original_df=df).reset_index(drop=True)

    def test_filter_rows_multi_condition(self):
        df = pd.DataFrame(
            [
                [100, None, '', 10],
                [250, 'brand1', False, np.NaN],
                [np.NaN, 'brand2', None, 18],
                [50, 'brand1', True, 13],
                [75, '', '', 80],
                [None, 'company3', False, 23],
            ],
            columns=['value', 'brand', 'discounted', 'inventory'],
        )
        action = dict(action_code='(value < 110 and value >= 50) and (value != null)')
        action2 = dict(action_code='brand contains brand and inventory != null')
        action3 = dict(action_code='(brand != null and value > 60) or (discounted == null)')
        action4 = dict(
            action_code='(discounted == True and inventory > 15)'
            ' or (discounted == False and value != null)'
        )
        action5 = dict(
            action_code='(brand not contains company and value == 75 and inventory <= 80)'
            ' or (discounted != null)'
        )
        df_expected = pd.DataFrame(
            [
                [100, None, '', 10],
                [50, 'brand1', True, 13],
                [75, '', '', 80],
            ],
            columns=['value', 'brand', 'discounted', 'inventory'],
        )
        df_expected2 = pd.DataFrame(
            [
                [np.NaN, 'brand2', None, 18],
                [50, 'brand1', True, 13],
            ],
            columns=['value', 'brand', 'discounted', 'inventory'],
        )
        df_expected3 = pd.DataFrame(
            [
                [100, None, '', 10],
                [250, 'brand1', False, np.NaN],
                [np.NaN, 'brand2', None, 18],
                [75, '', '', 80],
            ],
            columns=['value', 'brand', 'discounted', 'inventory'],
        )
        df_expected4 = pd.DataFrame(
            [
                [250, 'brand1', False, np.NaN],
            ],
            columns=['value', 'brand', 'discounted', 'inventory'],
        )
        df_expected5 = pd.DataFrame(
            [
                [250, 'brand1', False, np.NaN],
                [50, 'brand1', True, 13],
                [75, '', '', 80],
                [None, 'company3', False, 23],
            ],
            columns=['value', 'brand', 'discounted', 'inventory'],
        )
        df_new = filter_rows(df, action, original_df=df).reset_index(drop=True)
        df_new2 = filter_rows(df, action2, original_df=df).reset_index(drop=True)
        df_new3 = filter_rows(df, action3, original_df=df).reset_index(drop=True)
        df_new4 = filter_rows(df, action4, original_df=df).reset_index(drop=True)
        df_new5 = filter_rows(df, action5, original_df=df).reset_index(drop=True)
        df_new['value'] = df_new['value'].astype(int)
        df_new['inventory'] = df_new['inventory'].astype(int)
        df_new2['brand'] = df_new2['brand'].astype(str)
        df_new2['inventory'] = df_new2['inventory'].astype(int)
        df_new4['value'] = df_new4['value'].astype(int)
        df_new4['brand'] = df_new4['brand'].astype(str)
        df_new4['discounted'] = df_new4['discounted'].astype(bool)
        assert_frame_equal(df_expected, df_new)
        assert_frame_equal(df_expected2, df_new2)
        assert_frame_equal(df_expected3, df_new3)
        assert_frame_equal(df_expected4, df_new4)
        assert_frame_equal(df_expected5, df_new5)

    def test_filter_row_implicit_null(self):
        # tests that implicit null values in the transformed dataframe are still removed
        df = pd.DataFrame(
            [
                [100, None, '', 10],
                [250, 'brand1', False, np.NaN],
                [np.NaN, 'brand2', None, 18],
                [50, 'brand1', True, 13],
                [75, '', '', 80],
                [None, 'company3', False, 23],
            ],
            columns=['value', 'brand', 'discounted', 'inventory'],
        )
        action_payload = {
            'action_type': 'filter',
            'action_code': '%{1} != null',
            'action_arguments': [],
            'action_options': {},
            'axis': 'row',
            'action_variables': {
                '1': {
                    'id': 'value',
                    'type': 'feature',
                    'feature': {'column_type': 'number', 'uuid': 'value'},
                },
            },
            'outputs': [],
        }
        action = BaseAction(action_payload)
        df_new = action.execute(df).reset_index(drop=True)
        df_expected = pd.DataFrame(
            [
                [100, None, '', 10],
                [250, 'brand1', False, np.NaN],
                [50, 'brand1', True, 13],
                [75, '', '', 80],
            ],
            columns=['value', 'brand', 'discounted', 'inventory'],
        )
        df_new['value'] = df_new['value'].astype(int)
        assert_frame_equal(df_expected, df_new)

    def test_filter_row_dirty_columns(self):
        df = pd.DataFrame(
            [
                [100, None, '', 10],
                [250, 'brand1', False, np.NaN],
                [np.NaN, 'brand2', None, 18],
                [50, 'brand1', True, 13],
                [75, '', '', 80],
                [None, 'company3', False, 23],
            ],
            columns=['Val ue', 'bra  23423  nd', 'dis>>> ??cou nted', 'invVe nTory'],
        )
        action = dict(action_code='("Val ue" < 110 and "Val ue" >= 50) and ("Val ue" != null)')
        action2 = dict(action_code='"bra  23423  nd" contains brand and "invVe nTory" != null')
        action3 = dict(
            action_code='("bra  23423  nd" != null and "Val ue" > 60) or ("dis>>> ??cou nted" == null)'
        )
        action4 = dict(
            action_code='("dis>>> ??cou nted" == True and "invVe nTory" > 15)'
            ' or ("dis>>> ??cou nted" == False and "Val ue" != null)'
        )
        action5 = dict(
            action_code='("bra  23423  nd" not contains company and "Val ue" == 75 and "invVe nTory" <= 80) or ("dis>>> ??cou nted" != null)'
        )
        df_expected = pd.DataFrame(
            [
                [100, None, '', 10],
                [50, 'brand1', True, 13],
                [75, '', '', 80],
            ],
            columns=['Val ue', 'bra  23423  nd', 'dis>>> ??cou nted', 'invVe nTory'],
        )
        df_expected2 = pd.DataFrame(
            [
                [np.NaN, 'brand2', None, 18],
                [50, 'brand1', True, 13],
            ],
            columns=['Val ue', 'bra  23423  nd', 'dis>>> ??cou nted', 'invVe nTory'],
        )
        df_expected3 = pd.DataFrame(
            [
                [100, None, '', 10],
                [250, 'brand1', False, np.NaN],
                [np.NaN, 'brand2', None, 18],
                [75, '', '', 80],
            ],
            columns=['Val ue', 'bra  23423  nd', 'dis>>> ??cou nted', 'invVe nTory'],
        )
        df_expected4 = pd.DataFrame(
            [
                [250, 'brand1', False, np.NaN],
            ],
            columns=['Val ue', 'bra  23423  nd', 'dis>>> ??cou nted', 'invVe nTory'],
        )
        df_expected5 = pd.DataFrame(
            [
                [250, 'brand1', False, np.NaN],
                [50, 'brand1', True, 13],
                [75, '', '', 80],
                [None, 'company3', False, 23],
            ],
            columns=['Val ue', 'bra  23423  nd', 'dis>>> ??cou nted', 'invVe nTory'],
        )
        df_new = filter_rows(df, action, original_df=df).reset_index(drop=True)
        df_new2 = filter_rows(df, action2, original_df=df).reset_index(drop=True)
        df_new3 = filter_rows(df, action3, original_df=df).reset_index(drop=True)
        df_new4 = filter_rows(df, action4, original_df=df).reset_index(drop=True)
        df_new5 = filter_rows(df, action5, original_df=df).reset_index(drop=True)
        df_new['Val ue'] = df_new['Val ue'].astype(int)
        df_new['invVe nTory'] = df_new['invVe nTory'].astype(int)
        df_new2['bra  23423  nd'] = df_new2['bra  23423  nd'].astype(str)
        df_new2['invVe nTory'] = df_new2['invVe nTory'].astype(int)
        df_new4['Val ue'] = df_new4['Val ue'].astype(int)
        df_new4['bra  23423  nd'] = df_new4['bra  23423  nd'].astype(str)
        df_new4['dis>>> ??cou nted'] = df_new4['dis>>> ??cou nted'].astype(bool)
        assert_frame_equal(df_expected, df_new)
        assert_frame_equal(df_expected2, df_new2)
        assert_frame_equal(df_expected3, df_new3)
        assert_frame_equal(df_expected4, df_new4)
        assert_frame_equal(df_expected5, df_new5)

    def test_filter_row_comparing_two_dirty_cols(self):
        df = pd.DataFrame(
            [
                [23.5, 0.2342],
                [-9230.3, 98],
                [23.5, -0.32342],
                [432, 432],
                [34.2342, 0.082392],
                [23.5, np.nan],
                [None, 0.2342],
            ],
            columns=['e e e e e  e e e', ' kas22d fe ($)'],
        )
        action = dict(
            action_code='("e e e e e  e e e" != null and " kas22d fe ($)" != null) and "e e e e e  e e e" > " kas22d fe ($)"'
        )
        action2 = dict(
            action_code='("e e e e e  e e e" != null and " kas22d fe ($)" != null) and "e e e e e  e e e" <= " kas22d fe ($)"'
        )
        df_expected = pd.DataFrame(
            [
                [23.5, 0.2342],
                [23.5, -0.32342],
                [34.2342, 0.082392],
            ],
            columns=['e e e e e  e e e', ' kas22d fe ($)'],
        )
        df_expected2 = pd.DataFrame(
            [
                [-9230.3, 98],
                [432, 432],
            ],
            columns=['e e e e e  e e e', ' kas22d fe ($)'],
        )
        df_new = filter_rows(df, action, original_df=df).reset_index(drop=True)
        df_new2 = filter_rows(df, action2, original_df=df).reset_index(drop=True)
        df_new['e e e e e  e e e'] = df_new['e e e e e  e e e'].astype(float)
        df_new2['e e e e e  e e e'] = df_new2['e e e e e  e e e'].astype(float)
        df_new[' kas22d fe ($)'] = df_new[' kas22d fe ($)'].astype(float)
        df_new2[' kas22d fe ($)'] = df_new2[' kas22d fe ($)'].astype(int)
        assert_frame_equal(df_expected, df_new)
        assert_frame_equal(df_expected2, df_new2)

    def test_filter_row_dirty_columns_two(self):
        df = pd.DataFrame(
            [
                [np.NaN, False],
                ['sfc@mailnet.com', True],
                ['fdss@emailserver.net', True],
                ['fsdfsdfdsfdsf', False],
                ['xyz@mailnet.com', False],
                ['eeeeasdf', True],
            ],
            columns=['e e e e e e email', 'subs crip tion'],
        )
        action = dict(
            action_code='"e e e e e e email" not contains mailnet',
        )
        action2 = dict(
            action_code='"e e e e e e email" not contains \'mailnet\'',
        )
        action3 = dict(
            action_code='"e e e e e e email" not contains @',
        )
        action4 = dict(
            action_code='"e e e e e e email" not contains \'^e+\w\'',
        )
        action_invalid = dict(action_code='"subs crip tion" not contains False')
        df_new = filter_rows(df, action, original_df=df).reset_index(drop=True)
        df_new2 = filter_rows(df, action2, original_df=df).reset_index(drop=True)
        df_new3 = filter_rows(df, action3, original_df=df).reset_index(drop=True)
        df_new4 = filter_rows(df, action4, original_df=df).reset_index(drop=True)
        df_expected1 = pd.DataFrame(
            [
                [np.NaN, False],
                ['fdss@emailserver.net', True],
                ['fsdfsdfdsfdsf', False],
                ['eeeeasdf', True],
            ],
            columns=['e e e e e e email', 'subs crip tion'],
        )
        df_expected2 = pd.DataFrame(
            [[np.NaN, False], ['fsdfsdfdsfdsf', False], ['eeeeasdf', True]],
            columns=['e e e e e e email', 'subs crip tion'],
        )
        df_expected3 = pd.DataFrame(
            [
                [np.NaN, False],
                ['sfc@mailnet.com', True],
                ['fdss@emailserver.net', True],
                ['fsdfsdfdsfdsf', False],
                ['xyz@mailnet.com', False],
            ],
            columns=['e e e e e e email', 'subs crip tion'],
        )
        assert_frame_equal(df_new, df_expected1)
        assert_frame_equal(df_new2, df_expected1)
        assert_frame_equal(df_new3, df_expected2)
        assert_frame_equal(df_new4, df_expected3)

        with self.assertRaises(Exception):
            _ = filter_rows(df, action_invalid, original_df=df).reset_index(drop=True)

    def test_original_df_column_name_padding(self):
        # tests edge cases for when columns with the special prefixes "orig_" and "tf_" are given as input
        df = pd.DataFrame(
            [[0, 1, None], [1, 2, np.NaN], [np.NaN, 3, 4], [3, None, 5]],
            columns=['col', 'orig_col', 'tf_col'],
        )
        df_expected = pd.DataFrame(
            [
                [0, 1, None],
                [1, 2, np.NaN],
            ],
            columns=['col', 'orig_col', 'tf_col'],
        )
        action = dict(action_code='(col != null) and (orig_col != null)')
        df_new = filter_rows(df, action, original_df=df)
        df_new['col'] = df_new['col'].astype(int)
        df_new['orig_col'] = df_new['orig_col'].astype(int)
        assert_frame_equal(df_new, df_expected)

    def test_valid_columns_names(self):
        df = pd.DataFrame(
            [
                [
                    False,
                    False,
                    False,
                    False,
                ],
            ],
            columns=[
                '+=  -*&^%$! ?~  |<>',
                'this.is.an.operator',
                'are,commas,an,operator',
                'are()[]{}operators',
            ],
        )
        for column in df.columns:
            with self.assertRaises(Exception):
                action = dict(action_code=f'{column} == False')
                filter_rows(df, action, original_df=df)

        for column in df.columns:
            action = dict(action_code=f'"{column}" == False')
            filter_rows(df, action, original_df=df)

    def test_sort_rows(self):
        df = pd.DataFrame(
            [
                [0, False, 'a'],
                [1, True, 'b'],
                [1, True, 'c'],
                [0, True, 'd'],
            ],
            columns=[
                'integer',
                'boolean',
                'string',
            ],
        )

        test_cases = [
            (dict(action_arguments=['integer']), df.iloc[[0, 3, 1, 2]]),
            (
                dict(action_arguments=['integer'], action_options=dict(ascending=False)),
                df.iloc[[1, 2, 0, 3]],
            ),
            (dict(action_arguments=['string']), df.iloc[[0, 1, 2, 3]]),
            (
                dict(action_arguments=['string'], action_options=dict(ascending=False)),
                df.iloc[[3, 2, 1, 0]],
            ),
        ]

        for action, val in test_cases:
            self.assertTrue(sort_rows(df, action).equals(val))

    def test_sort_rows_with_multiple_columns(self):
        df = pd.DataFrame(
            [
                [0, False, 'a'],
                [1, True, 'b'],
                [1, True, 'c'],
                [0, True, 'd'],
            ],
            columns=[
                'integer',
                'boolean',
                'string',
            ],
        )

        test_cases = [
            (dict(action_arguments=['integer', 'string']), df.iloc[[0, 3, 1, 2]]),
            (
                dict(
                    action_arguments=['integer', 'string'],
                    action_options=dict(ascendings=[False, False]),
                ),
                df.iloc[[2, 1, 3, 0]],
            ),
            (
                dict(
                    action_arguments=['integer', 'string'],
                    action_options=dict(ascendings=[True, False]),
                ),
                df.iloc[[3, 0, 2, 1]],
            ),
            (
                dict(action_arguments=['string', 'integer'], action_options=dict(ascending=False)),
                df.iloc[[3, 2, 1, 0]],
            ),
        ]

        for action, val in test_cases:
            self.assertTrue(sort_rows(df, action).equals(val))

    def test_sort_rows_with_number_and_empty_strings(self):
        df = pd.DataFrame(
            [
                [0],
                [None],
                [3],
                [''],
                [1],
                [2],
            ],
            columns=[
                'integer',
            ],
        )

        test_cases = [
            (dict(ascending=True), df.iloc[[1, 3, 0, 4, 5, 2]]),
            (dict(ascending=False), df.iloc[[2, 5, 4, 0, 1, 3]]),
        ]
        for action_options, val in test_cases:
            action = dict(
                action_arguments=['integer'],
                action_variables={
                    '1': dict(
                        feature=dict(
                            column_type='number',
                            uuid='integer',
                        ),
                    ),
                },
                action_options=action_options,
            )

            self.assertTrue(sort_rows(df, action).equals(val))

    def test_remove_rows(self):
        df = pd.DataFrame(
            [
                [0, False, 'a'],
                [1, True, 'b'],
                [1, True, 'c'],
                [0, True, 'd'],
            ],
            columns=[
                'integer',
                'boolean',
                'string',
            ],
        )
        action1 = dict(
            action_type='remove',
            action_arguments=[],
            action_code='',
            action_options={
                'rows': [0, 3, 4, 5],
            },
            action_variables={},
            axis='row',
            outputs=[],
        )
        action2 = dict(
            action_type='remove',
            action_arguments=[],
            action_code='',
            action_options={
                'rows': [-1, 0, 1, 2, 3, 4],
            },
            action_variables={},
            axis='row',
            outputs=[],
        )
        expected_df = pd.DataFrame(
            [
                [1, True, 'b'],
                [1, True, 'c'],
            ],
            columns=[
                'integer',
                'boolean',
                'string',
            ],
        )
        expected_df.index = pd.Int64Index([1, 2])
        expected_df2 = pd.DataFrame(
            [],
            columns=[
                'integer',
                'boolean',
                'string',
            ],
        )
        expected_df2.index = pd.Int64Index([])
        new_df = remove_row(df, action1)
        new_df2 = remove_row(df, action2).astype(object)
        assert_frame_equal(new_df, expected_df)
        assert_frame_equal(new_df2, expected_df2)



from jupyter_client import KernelClient, KernelManager
from jupyter_client.kernelspec import NoSuchKernel
from mage_ai.server.kernels import DEFAULT_KERNEL_NAME, KernelName, kernel_managers


class ActiveKernel():
    def __init__(self):
        self.kernel = kernel_managers[DEFAULT_KERNEL_NAME]
        self.kernel_client = self.kernel.client()


active_kernel = ActiveKernel()


def switch_active_kernel(kernel_name: KernelName) -> None:
    if kernel_managers[kernel_name].is_alive():
        return

    for kernel in kernel_managers.values():
        if kernel.is_alive():
            kernel.request_shutdown()

    try:
        new_kernel = kernel_managers[kernel_name]
        new_kernel.start_kernel()
        active_kernel.kernel = new_kernel
        active_kernel.kernel_client = new_kernel.client()
    except NoSuchKernel as e:
        if kernel_name == KernelName.PYSPARK:
            raise Exception(
                'PySpark kernel is not installed. Please follow the instructions in '
                'https://github.com/mage-ai/mage-ai/blob/master/docs/kernels/README.md#pyspark-kernel-'
                'to install it.'
            ) from e
        else:
            raise e


def get_active_kernel() -> KernelManager:
    return active_kernel.kernel


def get_active_kernel_name() -> str:
    return active_kernel.kernel.kernel_name
 

def get_active_kernel_client() -> KernelClient:
    return active_kernel.kernel_client


def interrupt_kernel() -> None:
    active_kernel.kernel.interrupt_kernel()


def restart_kernel() -> None:
    active_kernel.kernel.restart_kernel()
    active_kernel.kernel_client = active_kernel.kernel.client()


def start_kernel() -> None:
    active_kernel.kernel.start_kernel()
    active_kernel.kernel_client = active_kernel.kernel.client()



from contextlib import contextmanager
from io import BytesIO
from mage_ai.io.base import BaseFile, FileFormat, QUERY_ROW_LIMIT
from mage_ai.io.config import BaseConfigLoader, ConfigKey
from pandas import DataFrame
from pathlib import Path
import boto3
import os


class S3(BaseFile):
    """
    Handles data transfer between a S3 bucket and the Mage app. Supports loading files
    of any of the following types:
    - ".csv"
    - ".json"
    - ".parquet"
    - ".hdf5"

    If profile is stored on file (in `~/.aws`), no further arguments are needed other than those
    specified below. Otherwise, use the factory method `with_credentials` to construct
    the data loader using manually specified credentials.
    """

    def __init__(
        self,
        verbose=False,
        **kwargs,
    ) -> None:
        """
        Initializes data loader from an S3 bucket. If IAM profile is stored on
        file (in `~/.aws`), no further arguments are needed other than those
        specified below. Otherwise, specify:
        - `aws_access_key_id` - AWS IAM access key ID
        - `aws_secret_access_key` - AWS IAM secret access key
        - `region_name` - name of AWS Region associated with profile
        """
        super().__init__(verbose=verbose)
        self.client = boto3.client('s3', **kwargs)

    def load(
        self,
        bucket_name: str,
        object_key: str,
        format: FileFormat = None,
        limit: int = QUERY_ROW_LIMIT,
        **kwargs,
    ) -> DataFrame:
        """
        Loads data from S3 into a Pandas data frame. This function will load at
        maximum 100,000 rows of data from the specified file.

        Args:
            import_config (Mapping, optional): Configuration settings for importing file from
            S3. Defaults to None.
            limit (int, Optional): The number of rows to limit the loaded dataframe to. Defaults to 100000.
            read_config (Mapping, optional): Configuration settings for reading file into data
            frame. Defaults to None.

        Returns:
            DataFrame: The data frame constructed from the file in the S3 bucket.
        """
        if format is None:
            format = self._get_file_format(object_key)
        with self.printer.print_msg(
            f'Loading data frame from bucket \'{bucket_name}\' at key \'{object_key}\''
        ):
            response = self.client.get_object(Bucket=bucket_name, Key=object_key)
        if format == FileFormat.HDF5:
            name = os.path.splitext(os.path.basename(object_key))[0]
            with self.open_temporary_directory() as temp_dir:
                obj_loc = temp_dir / f'{name}.hdf5'
                with obj_loc.open('wb') as fin:
                    fin.write(response['Body'].read())
                return self._read(obj_loc, format, limit, **kwargs)
        else:
            buffer = BytesIO(response['Body'].read())
            return self._read(buffer, format, limit, **kwargs)

    def export(
        self, df: DataFrame, bucket_name: str, object_key: str, format: FileFormat = None, **kwargs
    ) -> None:
        """
        Exports data frame to an S3 bucket.

        Args:
            df (DataFrame): Data frame to export
            write_config (Mapping, optional): Configuration settings for writing data frame to
            specified format. Defaults to None.
            export_config (Mapping, optional): Configuration settings for exporting data frame
            to S3. Defaults to None.
        """
        if format is None:
            format = self._get_file_format(object_key)

        with self.printer.print_msg(
            f'Exporting data frame to bucket \'{bucket_name}\' at key \'{object_key}\''
        ):
            if format == FileFormat.HDF5:
                name = os.path.splitext(os.path.basename(object_key))[0]
                with self.open_temporary_directory() as temp_dir:
                    obj_loc = temp_dir / f'{name}.hdf5'
                    self._write(df, format, obj_loc, **kwargs)
                    with obj_loc.open('rb') as fin:
                        self.client.put_object(Body=fin, Bucket=bucket_name, Key=object_key)
            else:
                buffer = BytesIO()
                self._write(df, format, buffer, **kwargs)
                buffer.seek(0)
                self.client.put_object(Body=buffer, Bucket=bucket_name, Key=object_key)

    @contextmanager
    def open_temporary_directory(self):
        temp_dir = Path.cwd() / '.tmp'
        temp_dir.mkdir(parents=True, exist_ok=True)
        yield temp_dir
        for file in temp_dir.iterdir():
            file.unlink()
        temp_dir.rmdir()

    @classmethod
    def with_config(
        cls,
        config: BaseConfigLoader,
        **kwargs,
    ) -> 'S3':
        """
        Initializes S3 client from configuration loader. This client accepts the following AWS
        IAM credential secrets:
        - Access Key ID
        - Secret Access Key
        - Region Name

        Args:
            config (BaseConfigLoader): Configuration loader object
        """
        return cls(
            aws_access_key_id=config[ConfigKey.AWS_ACCESS_KEY_ID],
            aws_secret_access_key=config[ConfigKey.AWS_SECRET_ACCESS_KEY],
            region_name=config[ConfigKey.AWS_REGION],
            **kwargs,
        )



from mage_ai.data_cleaner.column_types.column_type_detector import find_syntax_errors
from mage_ai.data_cleaner.column_types.constants import NUMBER_TYPES, ColumnType
from mage_ai.data_cleaner.shared.utils import clean_dataframe
from mage_ai.shared.constants import SAMPLE_SIZE
from mage_ai.shared.custom_types import FrozenDict
from mage_ai.shared.hash import merge_dict
from mage_ai.shared.logger import timer, VerboseFunctionExec
import math
import numpy as np
import pandas as pd
import logging


EMAIL_DOMAIN_REGEX = r'\@([^\s]*)'
INVALID_VALUE_SAMPLE_COUNT = 100
OUTLIER_SAMPLE_COUNT = 100
OUTLIER_ZSCORE_THRESHOLD = 3
PUNCTUATION = r'[:;\.,\/\\&`"\'\(\)\[\]\{\}]'
STOP_WORD_LIST = frozenset(['is', 'and', 'yet', 'but', 'a', 'or', 'nor', 'not', 'to', 'the'])
VALUE_COUNT_LIMIT = 20

logger = logging.getLogger(__name__)


def increment(metric, tags):
    pass


class StatisticsCalculator:
    """
    Key Assumption: statistics are clean - all values of the correct exact type at this point
    """

    def __init__(
        self,
        # s3_client,
        # object_key_prefix,
        # feature_set_version,
        column_types,
        verbose=False,
        **kwargs,
    ):
        self.column_types = column_types
        self.verbose = verbose

    @property
    def data_tags(self):
        return dict()

    def process(self, df, df_original=None, is_clean=True):
        return self.calculate_statistics_overview(df, df_original=df_original, is_clean=is_clean)

    def calculate_statistics_overview(self, df, df_original=None, is_clean=True):
        increment(
            'statistics.calculate_statistics_overview.start',
            self.data_tags,
        )
        with VerboseFunctionExec('Calculating statistics per variable', verbose=self.verbose):
            return self.__calculate_statistics_overview(df, df_original=df_original, is_clean=is_clean)

    def null_seq_gen(self, arr):
        prev = -1
        for is_null in arr:
            if is_null:
                prev += 1
            else:
                prev = 0
            yield prev

    def statistics_overview(self, series, col):
        try:
            return self.__statistics_overview(series, col)
        except Exception as err:
            increment(
                'statistics.calculate_statistics_overview.column.failed',
                merge_dict(
                    self.data_tags,
                    {
                        'col': col,
                        'error': err.__class__.__name__,
                    },
                ),
            )
            logger.exception(f'An error was caught while processing statistics: {err}')
            return {}

    def __calculate_statistics_overview(self, df, df_original=None, is_clean=True):
        with timer('statistics.calculate_statistics_overview.time', self.data_tags, verbose=False):
            if not is_clean:
                df = clean_dataframe(df, self.column_types, dropna=False)
            data = dict(
                original_row_count=len((df if df_original is None else df_original).index),
                count=len(df.index)
            )

            arr_args_1 = ([df[col] for col in df.columns],)
            arr_args_2 = ([col for col in df.columns],)

            dicts = map(self.statistics_overview, *arr_args_1, *arr_args_2)

            for d in dicts:
                data.update(d)

            # Aggregated stats
            column_count = len(df.columns)
            row_count = df.shape[0]

            data['total_null_value_count'] = sum(
                data[f'{col}/null_value_count'] for col in df.columns
            )
            data['total_null_value_rate'] = self.__protected_division(
                data['total_null_value_count'], df.size
            )
            data['total_invalid_value_count'] = sum(
                data[f'{col}/invalid_value_count'] for col in df.columns
            )
            data['total_invalid_value_rate'] = self.__protected_division(
                data['total_invalid_value_count'], df.size
            )

            df_dedupe = df.drop_duplicates()
            data['duplicate_row_count'] = row_count - df_dedupe.shape[0]

            data['empty_column_count'] = len(
                [col for col in df.columns if data[f'{col}/count'] == 0]
            )

            data['avg_null_value_count'] = self.__protected_division(
                data['total_null_value_count'], column_count
            )
            data['avg_invalid_value_count'] = self.__protected_division(
                data['total_invalid_value_count'], column_count
            )
            data['empty_column_rate'] = self.__protected_division(
                data['empty_column_count'], column_count
            )
            data['duplicate_row_rate'] = self.__protected_division(
                data['duplicate_row_count'], row_count
            )

            data['completeness'] = 1 - self.__protected_division(
                data['avg_null_value_count'], data['count']
            )
            data['validity'] = data['completeness'] - self.__protected_division(
                data['avg_invalid_value_count'], data['count']
            )

            timeseries_metadata = self.__evaluate_timeseries(data)
            data.update(timeseries_metadata)

            # object_key = s3_paths.path_statistics_overview(self.object_key_prefix)
            # s3_data.upload_json_sorted(self.s3_client, object_key, data)

        increment(
            'statistics.calculate_statistics_overview.success',
            self.data_tags,
        )

        return data

    def __evaluate_timeseries(self, data):
        indices = []
        for column, dtype in self.column_types.items():
            if data[f'{column}/null_value_rate'] <= 0.1 and dtype == ColumnType.DATETIME:
                indices.append(column)
        return {'is_timeseries': len(indices) != 0, 'timeseries_index': indices}

    def __protected_division(self, dividend: float, divisor: float) -> float:
        return dividend / divisor if divisor != 0 else 0

    def __statistics_overview(self, series, col):
        # The following regex based replace has high overheads
        # series = series.replace(r'^\s*$', np.nan, regex=True)
        df_value_counts = series.value_counts(dropna=False)

        df_top_value_counts = df_value_counts.copy()
        if df_top_value_counts.shape[0] > VALUE_COUNT_LIMIT:
            df_top_value_counts = df_top_value_counts.head(VALUE_COUNT_LIMIT)

        # TODO: remove duplicate data for distinct values
        # object_key_distinct_values = s3_paths.path_distinct_values_by_column(self.object_key_prefix, col)
        # s3_data.upload_dataframe(self.s3_client, df_top_value_counts, object_key_distinct_values, columns=[col])
        # object_key_statistics = s3_paths.path_statistics_by_column(self.object_key_prefix, col)
        # s3_data.upload_dataframe(self.s3_client, df_top_value_counts, object_key_statistics)

        # features = self.feature_set_version['features']
        # feature = find(lambda x: x['uuid'] == col, features)

        # if feature and feature.get('transformed'):
        #     return {}

        column_type = self.column_types.get(col)
        series_non_null = series.dropna()

        # Fix json serialization issue
        df_top_value_counts.index = pd.Index(str(idx) for idx in df_top_value_counts.index)

        df_value_counts_non_null = df_value_counts[df_value_counts.index.notnull()]
        count_unique = len(df_value_counts_non_null)

        data = {
            f'{col}/count': series_non_null.size,
            f'{col}/count_distinct': count_unique,
            f'{col}/null_value_count': series.isnull().sum(),
            f'{col}/value_counts': df_top_value_counts.to_dict(),
        }

        data[f'{col}/null_value_rate'] = self.__protected_division(
            data[f'{col}/null_value_count'], series.size
        )
        data[f'{col}/unique_value_rate'] = self.__protected_division(
            data[f'{col}/count_distinct'], series.size
        )
        data[f'{col}/max_null_seq'] = (
            max(self.null_seq_gen(series.isna().to_numpy()))
            if data[f'{col}/count'] != 0
            else len(series)
        )

        invalid_rows = find_syntax_errors(series_non_null, column_type)

        dates = None
        if len(series_non_null) > 0:
            if column_type in NUMBER_TYPES:
                data[f'{col}/average'] = series_non_null.mean()
                data[f'{col}/max'] = series_non_null.max()
                data[f'{col}/median'] = series_non_null.median()
                data[f'{col}/min'] = series_non_null.min()
                data[f'{col}/sum'] = series_non_null.sum()
                data[f'{col}/skew'] = series_non_null.skew()
                data[f'{col}/std'] = series_non_null.std()
                # detect outliers
                if data[f'{col}/std'] == 0:
                    data[f'{col}/outlier_count'] = 0
                else:
                    series_z_score = (
                        (series_non_null - data[f'{col}/average']) / data[f'{col}/std']
                    ).abs()
                    series_outliers = series_z_score[series_z_score >= OUTLIER_ZSCORE_THRESHOLD]
                    data[f'{col}/outlier_count'] = series_outliers.count()
                    data[f'{col}/outlier_ratio'] = self.__protected_division(
                        data[f'{col}/outlier_count'], series.size
                    )
                    data[f'{col}/outliers'] = (
                        series_non_null.loc[series_outliers.index]
                        .iloc[:OUTLIER_SAMPLE_COUNT]
                        .tolist()
                    )
                # generate five number summary
                first_quartile = series_non_null.quantile(0.25, interpolation='nearest')
                third_quartile = series_non_null.quantile(0.75, interpolation='nearest')
                iqr = third_quartile - first_quartile
                outlier_mask = (series_non_null <= first_quartile - 1.5 * iqr) | (
                    series_non_null >= third_quartile + 1.5 * iqr
                )
                outliers = series_non_null[outlier_mask].unique().tolist()
                data[f'{col}/box_plot_data'] = {
                    'outliers': outliers[:OUTLIER_SAMPLE_COUNT],
                    'min': data[f'{col}/min'],
                    'first_quartile': first_quartile,
                    'median': data[f'{col}/median'],
                    'third_quartile': third_quartile,
                    'max': data[f'{col}/max'],
                }
                if len(outliers) != 0:
                    not_outliers = series_non_null[~outlier_mask]
                    data[f'{col}/box_plot_data']['min'] = not_outliers.min()
                    data[f'{col}/box_plot_data']['max'] = not_outliers.max()
            elif column_type == ColumnType.DATETIME:
                dates = pd.to_datetime(series_non_null, utc=True, errors='coerce').dropna()
                data[f'{col}/max'] = dates.max().isoformat()
                data[f'{col}/median'] = (
                    dates.sort_values().iloc[math.floor(len(dates) / 2)].isoformat()
                )
                data[f'{col}/min'] = dates.min().isoformat()
            elif column_type == ColumnType.TEXT:
                text_series = series_non_null
                string_length = text_series.str.len()
                data[f'{col}/avg_string_length'] = string_length.mean()
                data[f'{col}/min_character_count'] = string_length.min()
                data[f'{col}/max_character_count'] = string_length.max()
                text_series = text_series.str.replace(PUNCTUATION, ' ', regex=True)
                text_series = text_series.str.lower().str.strip()
                text_series = text_series.str.split('\s+')

                word_count = text_series.map(len)
                data[f'{col}/max_word_count'] = word_count.max()
                data[f'{col}/avg_word_count'] = word_count.mean()
                data[f'{col}/min_word_count'] = word_count.min()

                exploded_text_series = text_series.explode()
                data[f'{col}/word_distribution'] = (
                    exploded_text_series.value_counts().head(VALUE_COUNT_LIMIT).to_dict()
                )
                # TODO: Calculate average word count excluding stopwords
                # data[f'{col}/word_count_excl_stopwords'] = (
                #     len(exploded_text_series) - exploded_text_series.isin(STOP_WORD_LIST).sum()
                # )
            elif column_type == ColumnType.EMAIL:
                valid_emails = series_non_null[~invalid_rows]
                domains = valid_emails.str.extract(EMAIL_DOMAIN_REGEX, expand=False)
                data[f'{col}/domain_distribution'] = (
                    domains.value_counts().head(VALUE_COUNT_LIMIT).to_dict()
                )
            elif column_type == ColumnType.LIST:
                lengths = series_non_null.str.len()
                data[f'{col}/avg_list_length'] = lengths.mean()
                data[f'{col}/max_list_length'] = lengths.max()
                data[f'{col}/min_list_length'] = lengths.min()
                data[f'{col}/length_distribution'] = (
                    lengths.value_counts().head(VALUE_COUNT_LIMIT).to_dict()
                )

                elements = series_non_null.explode()
                element_value_counts = elements.value_counts(dropna=False)
                data[f'{col}/most_frequent_element'] = element_value_counts.index[0]
                data[f'{col}/least_frequent_element'] = element_value_counts.index[-1]
                if any(
                    isinstance(idx, FrozenDict)
                    for idx in element_value_counts.index[:VALUE_COUNT_LIMIT]
                ):
                    str_index = element_value_counts.index.astype(str)
                    element_value_counts.index = str_index
                data[f'{col}/element_distribution'] = element_value_counts.head(
                    VALUE_COUNT_LIMIT
                ).to_dict()

        if column_type not in NUMBER_TYPES:
            if dates is not None:
                df_value_counts = dates.value_counts()
                string_df_value_counts = df_value_counts.head(VALUE_COUNT_LIMIT)
                string_df_value_counts.index = string_df_value_counts.index.astype(str)
                data[f'{col}/value_counts'] = string_df_value_counts.to_dict()

            mode, mode_idx = None, 0
            while mode_idx < count_unique and df_value_counts.index[mode_idx] in [None, np.nan]:
                mode_idx += 1
            if mode_idx < count_unique:
                mode = df_value_counts.index[mode_idx]

            if column_type == ColumnType.DATETIME and mode is not None:
                data[f'{col}/mode'] = mode.isoformat()
            else:
                data[f'{col}/mode'] = mode

            data[f'{col}/mode_ratio'] = (
                df_value_counts[mode].item() / df_value_counts.sum() if mode else 0
            )

        # Detect mismatched formats for some column types
        data[f'{col}/invalid_value_count'] = invalid_rows.sum()
        invalid_values = series_non_null[invalid_rows]
        data[f'{col}/invalid_values'] = invalid_values[:INVALID_VALUE_SAMPLE_COUNT].tolist()
        invalid_indices = series.index.get_indexer(invalid_values.index)
        data[f'{col}/invalid_indices'] = invalid_indices[
            np.where(invalid_indices <= SAMPLE_SIZE)
        ].tolist()
        data[f'{col}/invalid_value_distribution'] = (
            invalid_values.value_counts().head(VALUE_COUNT_LIMIT).to_dict()
        )
        data[f'{col}/invalid_value_rate'] = self.__protected_division(
            data[f'{col}/invalid_value_count'], series.size
        )

        # Calculate quality metrics
        data[f'{col}/completeness'] = 1 - data[f'{col}/null_value_rate']
        data[f'{col}/validity'] = data[f'{col}/completeness'] - data[f'{col}/invalid_value_rate']
        data[f'{col}/quality'] = 'Good' if data[f'{col}/validity'] >= 0.8 else 'Bad'

        return data



import setuptools


def readme():
    with open('README.md', encoding='utf8') as f:
        README = f.read()
    return README


requirements = []
with open('requirements.txt') as f:
    for line in f.read().splitlines():
        requirements.append(line)

setuptools.setup(
    name='mage-ai',
    version='0.2.3',
    author='Mage',
    author_email='eng@mage.ai',
    description='Mage is a notebook for building and deploying data pipelines.',
    long_description=readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/mage-ai/mage-ai',
    packages=setuptools.find_packages('./'),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    install_requires=requirements,
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'mage=mage_ai.command_line:main',
        ],
    },
    extras_require={},
)



from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.config import ConfigFileLoader
from mage_ai.io.redshift import Redshift
from pandas import DataFrame
from os import path

if 'data_exporter' not in globals():
    from mage_ai.data_preparation.decorators import data_exporter


@data_exporter
def export_data_to_redshift(df: DataFrame, **kwargs) -> None:
    """
    Template for exporting data to a Redshift cluster.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/blocks/data_loading.md#redshift
    """
    table_name = 'your_table_name'
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    with Redshift.with_config(ConfigFileLoader(config_path, config_profile)) as loader:
        loader.export(df, table_name)



from mage_ai.data_preparation.models.block import Block
from mage_ai.data_preparation.models.constants import BlockType, DATAFRAME_SAMPLE_COUNT_PREVIEW
from mage_ai.data_preparation.models.file import File
from mage_ai.data_preparation.models.pipeline import Pipeline
from mage_ai.data_preparation.repo_manager import get_repo_path, init_repo, set_repo_path
from mage_ai.data_preparation.utils.block.convert_content import convert_to_block
from mage_ai.data_preparation.variable_manager import VariableManager
from mage_ai.server.active_kernel import (
    interrupt_kernel,
    restart_kernel,
    start_kernel,
    switch_active_kernel,
)
from mage_ai.server.api.autocomplete_items import ApiAutocompleteItemsHandler
from mage_ai.server.api.base import BaseHandler
from mage_ai.server.api.widgets import ApiPipelineWidgetDetailHandler, ApiPipelineWidgetListHandler
from mage_ai.server.constants import DATA_PREP_SERVER_PORT
from mage_ai.server.kernel_output_parser import parse_output_message
from mage_ai.server.kernels import (
    DEFAULT_KERNEL_NAME,
    kernel_managers,
    KernelName,
    PIPELINE_TO_KERNEL_NAME,
)
from mage_ai.server.subscriber import get_messages
from mage_ai.server.websocket import WebSocketServer
import argparse
import asyncio
import json
import os
import socket
import tornado.ioloop
import tornado.web
import urllib.parse


class MainHandler(tornado.web.RequestHandler):
    def get(self, *args):
        self.render('index.html')


class ApiBlockHandler(BaseHandler):
    def delete(self, block_type_and_uuid_encoded):
        block_type_and_uuid = urllib.parse.unquote(block_type_and_uuid_encoded)
        parts = block_type_and_uuid.split('/')
        if len(parts) != 2:
            raise Exception('The url path should be in block_type/block_uuid format.')
        block_type = parts[0]
        block_uuid = parts[1]
        block = Block(block_uuid, block_uuid, block_type)
        if not block.exists():
            raise Exception(f'Block {block_uuid} does not exist')
        block.delete()
        self.write(dict(block=block.to_dict()))


class ApiFileListHandler(BaseHandler):
    def get(self):
        self.write(dict(files=[File.get_all_files(get_repo_path())]))

    def post(self):
        data = json.loads(self.request.body).get('file', {})
        file = File.create(data.get('name'), data.get('dir_path'), get_repo_path())
        self.write(dict(file=file.to_dict()))


class ApiFileContentHandler(BaseHandler):
    def get(self, file_path_encoded):
        file_path = urllib.parse.unquote(file_path_encoded)
        file = File.from_path(file_path, get_repo_path())

        self.write(dict(file_content=file.to_dict(include_content=True)))

    def put(self, file_path_encoded):
        file_path = urllib.parse.unquote(file_path_encoded)
        file = File.from_path(file_path, get_repo_path())

        data = json.loads(self.request.body).get('file_content', {})
        content = data.get('content')
        if content is None:
            raise Exception('Please provide a \'content\' param in the request payload.')
        file.update_content(content)

        self.write(dict(file_content=file.to_dict(include_content=True)))


class ApiPipelineHandler(BaseHandler):
    def delete(self, pipeline_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        response = dict(pipeline=pipeline.to_dict())
        pipeline.delete()
        self.write(response)

    def get(self, pipeline_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        include_content = self.get_bool_argument('include_content', True)
        include_outputs = self.get_bool_argument('include_outputs', True)
        switch_active_kernel(PIPELINE_TO_KERNEL_NAME[pipeline.type])
        self.write(
            dict(
                pipeline=pipeline.to_dict(
                    include_content=include_content,
                    include_outputs=include_outputs,
                    sample_count=DATAFRAME_SAMPLE_COUNT_PREVIEW,
                )
            )
        )
        self.finish()

    def put(self, pipeline_uuid):
        """
        Allow updating pipeline name and uuid
        """
        pipeline = Pipeline.get(pipeline_uuid)
        update_content = self.get_bool_argument('update_content', False)
        data = json.loads(self.request.body).get('pipeline', {})
        pipeline.update(data, update_content=update_content)
        switch_active_kernel(PIPELINE_TO_KERNEL_NAME[pipeline.type])
        self.write(
            dict(
                pipeline=pipeline.to_dict(
                    include_content=update_content,
                    include_outputs=update_content,
                    sample_count=DATAFRAME_SAMPLE_COUNT_PREVIEW,
                )
            )
        )


class ApiPipelineExecuteHandler(BaseHandler):
    def post(self, pipeline_uuid):
        pipeline = Pipeline.get(pipeline_uuid)

        global_vars = None
        if len(self.request.body) != 0:
            global_vars = json.loads(self.request.body).get('global_vars')

        asyncio.run(pipeline.execute(global_vars=global_vars))
        self.write(
            dict(
                pipeline=pipeline.to_dict(
                    include_outputs=True,
                    sample_count=DATAFRAME_SAMPLE_COUNT_PREVIEW,
                )
            )
        )
        self.finish()


class ApiPipelineListHandler(BaseHandler):
    def get(self):
        pipelines = Pipeline.get_all_pipelines(get_repo_path())
        self.write(dict(pipelines=pipelines))
        self.finish()

    def post(self):
        pipeline = json.loads(self.request.body).get('pipeline', {})
        clone_pipeline_uuid = pipeline.get('clone_pipeline_uuid')
        name = pipeline.get('name')
        if clone_pipeline_uuid is None:
            pipeline = Pipeline.create(name, get_repo_path())
        else:
            source = Pipeline.get(clone_pipeline_uuid)
            pipeline = Pipeline.duplicate(source, name)
        self.write(dict(pipeline=pipeline.to_dict()))


class ApiPipelineBlockHandler(BaseHandler):
    def get(self, pipeline_uuid, block_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        block = pipeline.get_block(block_uuid)
        include_outputs = self.get_bool_argument('include_outputs', True)
        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
        self.write(
            dict(
                block=block.to_dict(
                    include_content=True,
                    include_outputs=include_outputs,
                )
            )
        )
        self.finish()

    def put(self, pipeline_uuid, block_uuid):
        """
        Allow updating block name, uuid, type, upstream_block, and downstream_blocks
        """
        pipeline = Pipeline.get(pipeline_uuid)
        data = json.loads(self.request.body).get('block', {})
        block = pipeline.get_block(block_uuid)
        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
        block.update(data)
        self.write(dict(block=block.to_dict(include_content=True)))

    def delete(self, pipeline_uuid, block_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        block = pipeline.get_block(block_uuid)
        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
        block.delete()
        self.write(dict(block=block.to_dict()))


class ApiPipelineBlockExecuteHandler(BaseHandler):
    def post(self, pipeline_uuid, block_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        block = pipeline.get_block(block_uuid)
        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
        asyncio.run(block.execute(redirect_outputs=True))
        self.write(
            dict(
                block=block.to_dict(
                    include_outputs=True,
                )
            )
        )
        self.finish()


class ApiPipelineBlockListHandler(BaseHandler):
    def get(self, pipeline_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        include_outputs = self.get_bool_argument('include_outputs', True)
        self.write(
            dict(
                blocks=pipeline.to_dict(
                    include_content=True,
                    include_outputs=include_outputs,
                    sample_count=DATAFRAME_SAMPLE_COUNT_PREVIEW,
                )['blocks']
            )
        )
        self.finish()

    def post(self, pipeline_uuid):
        """
        Create block and add to pipeline
        """
        pipeline = Pipeline.get(pipeline_uuid)
        payload = json.loads(self.request.body).get('block', {})
        block = Block.create(
            payload.get('name') or payload.get('uuid'),
            payload.get('type'),
            get_repo_path(),
            config=payload.get('config'),
            pipeline=pipeline,
            priority=payload.get('priority'),
            upstream_block_uuids=payload.get('upstream_blocks', []),
        )

        content = payload.get('content')
        if content:
            if payload.get('converted_from'):
                content = convert_to_block(block, content)

            block.update_content(content)

        pipeline.add_block(block, payload.get('upstream_blocks', []))
        self.write(dict(block=block.to_dict(include_content=True)))


class ApiPipelineBlockAnalysisHandler(BaseHandler):
    def get(self, pipeline_uuid, block_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        block = pipeline.get_block(block_uuid)
        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
        analyses = block.get_analyses()
        self.write(dict(analyses=analyses))


class ApiPipelineBlockOutputHandler(BaseHandler):
    def get(self, pipeline_uuid, block_uuid):
        pipeline = Pipeline.get(pipeline_uuid)
        block = pipeline.get_block(block_uuid)
        if block is None:
            raise Exception(f'Block {block_uuid} does not exist in pipeline {pipeline_uuid}')
        outputs = block.get_outputs(sample_count=None)
        self.write(dict(outputs=outputs))


class ApiPipelineVariableListHandler(BaseHandler):
    def get(self, pipeline_uuid):
        variables_dict = VariableManager(get_repo_path()).get_variables_by_pipeline(pipeline_uuid)
        variables = [
            dict(
                block=dict(uuid=uuid),
                pipeline=dict(uuid=pipeline_uuid),
                variables=arr,
            )
            for uuid, arr in variables_dict.items()
        ]
        self.write(dict(variables=variables))
        self.finish()


class KernelsHandler(BaseHandler):
    def get(self, kernel_id=None):
        kernels = []

        for kernel_name in KernelName:
            kernel = kernel_managers[kernel_name]
            if kernel.has_kernel:
                kernels.append(
                    dict(
                        alive=kernel.is_alive(),
                        id=kernel.kernel_id,
                        name=kernel.kernel_name,
                    )
                )

        r = json.dumps(dict(kernels=kernels))
        self.write(r)

    def post(self, kernel_id, action_type):
        kernel_name = self.get_argument('kernel_name', DEFAULT_KERNEL_NAME)
        if kernel_name not in kernel_managers:
            kernel_name = DEFAULT_KERNEL_NAME
        switch_active_kernel(kernel_name)
        if 'interrupt' == action_type:
            interrupt_kernel()
        elif 'restart' == action_type:
            try:
                restart_kernel()
            except RuntimeError as e:
                # RuntimeError: Cannot restart the kernel. No previous call to 'start_kernel'.
                if 'start_kernel' in str(e):
                    start_kernel()

        r = json.dumps(
            dict(
                kernel=dict(
                    id=kernel_id,
                ),
            )
        )
        self.write(r)
        self.finish()


def make_app():
    routes = [
        (r'/', MainHandler),
        # (r'/pipelines', MainHandler),
        (r'/pipelines/(.*)', MainHandler),
        (
            r'/_next/static/(.*)',
            tornado.web.StaticFileHandler,
            {'path': os.path.join(os.path.dirname(__file__), 'frontend_dist/_next/static')},
        ),
        (
            r'/fonts/(.*)',
            tornado.web.StaticFileHandler,
            {'path': os.path.join(os.path.dirname(__file__), 'frontend_dist/fonts')},
        ),
        (
            r'/(favicon.ico)',
            tornado.web.StaticFileHandler,
            {'path': os.path.join(os.path.dirname(__file__), 'frontend_dist')},
        ),
        (r'/websocket/', WebSocketServer),
        (r'/api/blocks/(?P<block_type_and_uuid_encoded>.+)', ApiBlockHandler),
        (r'/api/files', ApiFileListHandler),
        (r'/api/file_contents/(?P<file_path_encoded>.+)', ApiFileContentHandler),
        (r'/api/pipelines/(?P<pipeline_uuid>\w+)/execute', ApiPipelineExecuteHandler),
        (r'/api/pipelines/(?P<pipeline_uuid>\w+)', ApiPipelineHandler),
        (r'/api/pipelines', ApiPipelineListHandler),
        (
            r'/api/pipelines/(?P<pipeline_uuid>\w+)/blocks/(?P<block_uuid>\w+)/execute',
            ApiPipelineBlockExecuteHandler,
        ),
        (
            r'/api/pipelines/(?P<pipeline_uuid>\w+)/blocks/(?P<block_uuid>\w+)',
            ApiPipelineBlockHandler,
        ),
        (
            r'/api/pipelines/(?P<pipeline_uuid>\w+)/blocks/(?P<block_uuid>\w+)/analyses',
            ApiPipelineBlockAnalysisHandler,
        ),
        (
            r'/api/pipelines/(?P<pipeline_uuid>\w+)/blocks/(?P<block_uuid>\w+)/outputs',
            ApiPipelineBlockOutputHandler,
        ),
        (r'/api/pipelines/(?P<pipeline_uuid>\w+)/blocks', ApiPipelineBlockListHandler),
        (r'/api/pipelines/(?P<pipeline_uuid>\w+)/variables', ApiPipelineVariableListHandler),
        (
            r'/api/pipelines/(?P<pipeline_uuid>\w+)/widgets/(?P<block_uuid>\w+)',
            ApiPipelineWidgetDetailHandler,
        ),
        (
            r'/api/pipelines/(?P<pipeline_uuid>\w+)/widgets',
            ApiPipelineWidgetListHandler,
        ),
        (r'/api/kernels', KernelsHandler),
        (r'/api/kernels/(?P<kernel_id>[\w\-]*)/(?P<action_type>[\w\-]*)', KernelsHandler),
        (r'/api/autocomplete_items', ApiAutocompleteItemsHandler),
    ]
    return tornado.web.Application(
        routes,
        autoreload=True,
        template_path=os.path.join(os.path.dirname(__file__), 'frontend_dist'),
    )


async def main(
    host: str = None,
    port: str = None,
    project: str = None,
):
    host = host if host else None
    port = port if port else DATA_PREP_SERVER_PORT
    project = project if project else None

    if project:
        project = project = os.path.abspath(project)
    else:
        project = os.path.join(os.getcwd(), 'default_repo')

    if not os.path.exists(project):
        init_repo(project)
    set_repo_path(project)

    switch_active_kernel(DEFAULT_KERNEL_NAME)

    app = make_app()

    def is_port_in_use(port: int) -> bool:
        print(f'Checking port {port}...')
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex(('localhost', port)) == 0

    port = int(port)
    max_port = port + 100
    while is_port_in_use(port):
        if port > max_port:
            raise Exception(
                'Unable to find an open port, please clear your running processes if possible.'
            )
        port += 1

    app.listen(
        port,
        address=host,
    )

    print(f'Mage is running at http://{host or "localhost"}:{port} and serving project {project}')

    get_messages(
        lambda content: WebSocketServer.send_message(
            parse_output_message(content),
        ),
    )

    await asyncio.Event().wait()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--host', type=str, default=None)
    parser.add_argument('--port', type=str, default=None)
    parser.add_argument('--project', type=str, default=None)
    args = parser.parse_args()

    host = args.host
    port = args.port
    project = args.project

    asyncio.run(
        main(
            host=host,
            port=port,
            project=project,
        )
    )



{% extends "transformers/transformer_actions/action.jinja" %}
{% block action %}
    """
    Execute Transformer Action: ActionType.SORT

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/actions/transformer_actions/README.md#sort
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.SORT,
        arguments=[],  # Specify columns to sort rows by
        axis=Axis.ROW,
        options={'ascending': True},
    )
{% endblock %}



{% extends "transformers/transformer_actions/action.jinja" %}
{% block action %}
    """
    Execute Transformer Action: ActionType.SHIFT_UP

    Shifts value in the selected column down by specified number periods.

    Docs: https://github.com/mage-ai/mage-ai/blob/master/docs/actions/transformer_actions/README.md#shift-up
    """
    action = build_transformer_action(
        df,
        action_type=ActionType.SHIFT_UP,
        arguments=[],  # Specify one column to perform shift on
        axis=Axis.COLUMN,
        outputs=[{'uuid': 'up_shifted_column_1', 'type': 'category'}],
    )
{% endblock %}
